class Statement;

class Assembler {
  u1* memory;
  int size, capacity;
  u2  currentClass;

  void patch2(int addr, u2 value) {
    assert(addr+1 < size && addr >= 0);
    memory[addr+0] = value >> 8;
    memory[addr+1] = value & 255;
  }

  void patch4(int addr, u4 value) {
    //    cerr << "at " << addr << " patch " << value << endl;

    patch2(addr+0, value >> 16);
    patch2(addr+2, value & 65535ul);
  }

  int _add1(u1 byte) {
    if(size == capacity) reserve(2 * capacity);
    int result = size++;
    memory[result] = byte;
    return result;
  }

public:
  static Statement* pre_ctor; // kludge

  void reserve(int new_capacity) {
    if(new_capacity > capacity) {
      u1* old_memory = memory;
      u1* new_memory = new u1[size];
      assert(new_memory);
      if(old_memory) memcpy(new_memory, old_memory, size);
      capacity = new_capacity;
      memory = new_memory;
      delete[] old_memory;
    }
  }

  explicit Assembler(int init = 16384):
    memory(0),capacity(0),size(0),codeAttr(0){ reserve(init); }
  ~Assembler(){delete[] memory;}
  
  int add2(u2 value, char* comment = 0) {
    int current = _add1(value >> 8);
    _add1(value & 255);
    //if(comment) cerr << comment << ": " << value << endl;
    return current;
  }

  int currentAddress() { return size; }

  int add4(u4 value) {
    int current = add2(value >> 16);
    add2(value & 65535u);
    return current;
  }

  void beginClass(u2 attr, u2 this_class, u2 super_class) { 
    size = 0; 
    
    add2(attr);
    add2(this_class); currentClass = this_class;
    add2(super_class); 
    
    //    cerr << attr << ' ' << this_class << ' ' << super_class << ' ';
    //    printMod(cerr, attr);  cpool.printClass(cerr, this_class);
    //    cerr << ' '; cpool.printClass(cerr, super_class); cerr << endl;
  }

  void beginField(u2 attr, u2 name, u2 type) {
    add2(attr);
    add2(name);
    add2(type);
    
    //    cpool.printUtf(cerr, name); 
    //   cerr << ' '; cpool.printUtf(cerr, type); cerr << ' '; printMod(cerr, attr); cerr << endl;
  }

  void endClass() {
    char buf[256];
    strcpy(buf, cpool.getClassName(currentClass));
    strcat(buf, ".class");
    ofstream o(buf);
    assert(!o.bad());
    static unsigned char start[] = {0xCA, 0xFE, 0xBA, 0xBE, 0,3, 0,45};
    cout << "writing " << buf << " ... " << flush;
    o.write(start, sizeof(start));
    cpool.write(o);
    o.write(memory, size);
    cout << "done" << endl;
  }
  //  int newLocal(int size) {
  //    int slot = max_locals;
  //    max_locals += size; 
  //  }
private:
  int codeAttr;
  Type* returnType; 
  u2 max_stack;
  u2 max_locals;
  u2 current_stack;
  u2 current_local;

  Type* locals[MAX_LOCALS]; 
public:
  int addLocal(Type* t) {
    assert(current_local < MAX_LOCALS);
    int slot = current_local;
    current_local += t->stackSize();
    locals[slot] = t;
    if(current_local > max_locals) max_locals = current_local;
    return slot;
  }
  void removeLocals(int n) {
    assert(n <= current_local);
    current_local -= n;
  }
  u2 mname;
  int maddr;
  void beginMethod(u2 attr, u2 name, Type* type) {
    assert(codeAttr == 0); 
    beginField(attr, name, type->no());
    add2(1, "Code Attribute");

    returnType = type->getReturnType(); 
    current_local = current_stack = max_stack = max_locals = 0;

    add2( Identifier::Code.no() );
    codeAttr = size;
    add4( 0 ); // attribute length - should be patched
    add4( 0 ); // max_stack, max_locals - should be patched
    add4( 0 ); // code size - should be patched
    mname = name; maddr = currentAddress();
  }
  Assembler& add(int opcode, int stack_before=-1, int stack_after=0) {
    assert(opcode >= 0 && opcode <= last_instruction);
    ins_desc* ins = ins_descriptions + opcode;

    if(stack_before == -1) {
      stack_before = ins->stackBefore;
      stack_after = ins->stackAfter;
    }
    
    if( current_stack < stack_before ) {
      cerr << "Complain: " << ins->name
	   << " requires " << stack_before 
	   << " word(s). Current stack size = " << current_stack 
	   << " " << cpool.getUtf(mname)  << ':' << size-maddr << endl;
      
      current_stack = stack_before;
    }
    current_stack += stack_after - stack_before;
    if(current_stack > max_stack) max_stack = current_stack;
    //    cerr << setw(2) << size-codeAttr-12 << ' ' << ins->name << endl; 
    //    if(ins->len == 0) cerr << endl;
    _add1(opcode);
    return *this;
  }
  void pad(int align) {
    int addr = size-codeAttr-12;
    int misa = addr % align;
    switch(misa) {
      case 0: return;
      case 1: _add1(0);
      case 2: _add1(0);
      case 3: _add1(0);
    } 
  }
  void ref(u2 no) {
    add2(no);
    //    cpool.printRef(cerr, no); cerr << endl;
  }
  void arrayCode(Type* t) {
    _add1(t->arrayCode());
    //    cerr << t->typeSig.asString()[0] << ' ' << t->arrayCode() << endl;
  }
  Assembler& type(Type* t) {
    add2(t->no());
    //    cerr << t;
    return *this;
  }
  void class_type(Type* t) {
    add2(t->classNo());
    //    cerr << t;
  }
  void byte(u1 value) {
    _add1(value);
    //    cerr << ' ' << (int)value << endl;
  }
  void iinc(u1 slot, int value) {
    assert(-128 <= value && value < 128);
    _add1(slot);
    _add1(u1(value));
    //    cerr << " local_" << int(slot) << ' ' << value << endl;
  }
  void word(u2 value) {
    add2(value);
    //    cerr << ' ' << (int)value << endl;
  }
  int _jump_ofs(int value) {
    int result =  add2(unsigned(value));
    //    cerr << ' '; if(value >= 0) cerr << '+'; 
    //    cerr << value << endl;
    return result;
  }
  int branch(int op, int value) {
    int result = currentAddress(); 
    add(op); _jump_ofs(value?value - result:0);
    return result;
  }
  void patch_branch(int addr, int value) { patch2(addr+1, value - addr);  }
  void patch_branch(int addr) { patch_branch(addr, currentAddress()); }
  void local(int op, int slot) { // op should be either op_iload or op_istore
    assert(slot >= 0 && slot < current_local);
    int offs = locals[slot]->offset();
    if(slot < 4) add(op + 5 + offs * 4 + slot);
    else { add(op + offs); _add1(slot); }
  }
  int load_int_const(int value) {
    int result = currentAddress();
    if(value >= -1 && value <= 5) add(op_iconst_0 + value);
    else if(-128 <= value && value < 128) { add(op_bipush).byte(value); }
    else if(-32768 <= value && value < 32768) { add(op_sipush); word(value); }
    else {
      u2 ndx = cpool.addInteger(value);
      if(ndx < 256) { add(op_ldc1); byte(ndx); }
      else {add(op_ldc2); add2(ndx); }
    }
    return result;
  }
  void checkStack() {
    if(current_stack != 0) {
      srcPos(cerr) << "Complain: stack size = " << current_stack 
	   << " (OK if comparison=>boolean)" << endl;
      current_stack = 0;
    }
  }
  void endMethod() {
    assert(codeAttr); //cerr << "codeAttr = " << codeAttr << endl;
    if(size == codeAttr + 12 || !isReturn(memory[size-1])) 
      add(op_ireturn + returnType->voffset() );
    checkStack();
    if(current_local != 0) {
      cerr << "Complain: current_local = " << current_local << endl;
    }
    //    cerr << "max_stack = " << max_stack << " max_locals = " 
    //	 << max_locals << endl;
    patch2(codeAttr + 4, max_stack);
    patch2(codeAttr + 6, max_locals);
    patch4(codeAttr + 8, size - codeAttr - 12);
    add2( 0 ); // exception table
    add2( 0 ); // attributes_count
    patch4(codeAttr + 0, size - codeAttr - 4);
    codeAttr = 0;
  }
};
Statement* Assembler::pre_ctor; // kludge
Assembler code;
