/************************************************
  Team: Jon Gregory, Gor Nishanov
  Project #2 Part 3
    for RPI course 66-648 Compiler Design Theory

  consts.h -- Here we declared a lot of constants;

  Also here there is a useful functions that
  print modifiers from a modifier mask:

    ostream& printMod(ostream& o, int modifiers);

    */  


extern int lineno;
char* sourceFileName = "<null>";

inline ostream& srcPos(ostream& o) {
  o << sourceFileName;
  if(lineno != -1) o << ':' << lineno;
  o << ": ";
  return o;
}

#define ASSERT1(cond,code) if(!(cond)){ srcPos(cerr) << # code << endl; abort();} else
#define ASSERT(cond,code,param) if(!(cond)){\
  srcPos(cerr) << # code << ": " << param << endl; abort();} else
#define ASSERT2(cond,code,p1,p2) if(!(cond)){\
    srcPos(cerr) << # code << ": " << p1 << ' ' << p2 << endl; abort();} else

struct Buf {
  enum{len = 16384};
private:
  int memory[len/sizeof(int)];
  Buf(const Buf&) {assert(false);}
  bool free;
  int no;
public:
  explicit Buf(int i,int):free(1),no(i){
    //    cerr << "Buf " << no << " initialized" << endl;
  }
  char* mem() {
    ASSERT(free, NOT_FREE_BUF, no); 
    free = 0;
    return (char*)memory;
  }
  void release() {free = 1;}
};

Buf buf1(1,0), buf2(2,0), buf3(3,0), buf4(4,0), buf(5,0);

const int TYPE_HASH_SIZE = 193;
const int PKGS_HASH_SIZE = 193;
const int NAME_HASH_SIZE = 389;
const int LOCL_HEAP_SIZE = 1024; // SymbolTable size in Kbytes

const int MAX_LOCALS = 256;

const char SIG_BYTE     = 'B'; 
const char SIG_CHAR     = 'C';
const char SIG_DOUBLE   = 'D';
const char SIG_FLOAT    = 'F';
const char SIG_INT      = 'I';
const char SIG_LONG     = 'J';
const char SIG_CLASSBEG = 'L';
const char SIG_CLASSEND = ';';
const char SIG_SHORT    = 'S';
const char SIG_BOOL     = 'Z';
const char SIG_VOID     = 'V';
const char SIG_ARRAY    = '[';
const char SIG_PARBEG   = '(';
const char SIG_PAREND   = ')';
 // double d[][][] == [[[D
 // Object mymethod(int i, double d, Thread t)
 // == (IDLjava/lang/Thread;)Ljava/lang/Object

const int ACC_PUBLIC    = 0x0001;
const int ACC_PRIVATE   = 0x0002;
const int ACC_PROTECTED = 0x0004;
const int ACC_STATIC    = 0x0008;
const int ACC_FINAL     = 0x0010;
const int ACC_SUPER     = 0x0020;
const int ACC_VOLATILE  = 0x0040;
const int ACC_TRANSIENT = 0x0080;
const int ACC_NATIVE    = 0x0100;
const int ACC_INTERFACE = 0x0200;
const int ACC_ABSTRACT  = 0x0400;
const int ACC_SYNCHRONIZED = 0x0800;
//const int ACC_CTOR      = 0x1000;

const char* typeNames[] = {
"Bool",  //const int TC_BOOL      = 0;
"Byte",  //const int TC_BYTE      = 1;
"Char",  //const int TC_CHAR      = 2;
"Short", //const int TC_SHORT     = 3;
"Int",   //const int TC_INT       = 4;
"Long",  //const int TC_LONG      = 5;
"Float", //const int TC_FLOAT     = 6;
"Double",//const int TC_DOUBLE    = 7;
"Null",  //const int TC_NULL      = 8;
"Array", //const int TC_ARRAY     = 9;
"Class", //const int TC_CLASS     = 10;
"Void",  //const int TC_VOID      = 11;
"Method",//const int TC_METHOD    = 12;
"Error"  //const int TC_ERROR     = 13;
};
const char* arrNames[] = {
  "Bool", // case TC_BOOL: return 4;
  "Char", //  case TC_CHAR: return 5;
  "Float", //  case TC_FLOAT: return 6;
  "Double", //  case TC_DOUBLE:return 7;
  "Byte", //  case TC_BYTE:  return 8;
  "Short", //  case TC_SHORT: return 9;
  "Int", //  case TC_INT: return 10;
  "Long", //  case TC_LONG: return 11;
};

inline void printArrCode(ostream& o, int typ) {
  if(typ < 4 || typ > 11) o << "BAD_TYPE_" << typ;
  else o << arrNames[typ - 4];
}

#define null NULL
const char* opNames[] = {
	",",		"=",		"+=",		"-=",		"*=",
	"/=",		"%=",		"<<=",		">>=",		"<<<=",
	"&=",		"|=",		"^=",		"?:",		"||",
	"&&",		"&",            "|",		"^",		"!=",
	"==", 		">=",		">",		"<=",		"<",
	"instanceof",	"<<",		">>",		"<<<",		"+",
	"-",		"/",		"%",		"*",		"cast",
	"+",		"-",		"!",		"~",		"++",
	"--",		"new",		"new",		"new",		"++",
	"--",		"field",	"method",	"[]",		"new",
	"++",		"--",		null,		null,		null,

	"convert",	"expr",		"array",	"goto",		null,

	"Identifier",	"boolean",	"byte",		"char",		"short",
	"int",		"long",		"float",	"double",	"string",

	"byte",		"char",		"short",	"int",		"long",
	"float",	"double",	"void",		"boolean",	null,

	"true",		"false",	"this",		"super",	"null",
	null,		null,		null,		null,		null,

	"if",		"else",		"for",		"while",	"do",
	"switch",	"case",		"default",	"break",	"continue",
	"return",	"try",		"catch",	"finally",	"throw",
	"stat",		"expression",	"declaration",	"declaration",  null,

	"import",	"class",	"extends",	"implements",	"interface",
	"package",	null,		null,		null,		null,

	"private",	"public",	"protected",	"const",	"static",
	"transient",	"synchronized",	"native",	"final",	"volatile",
	"abstract",	null,		null,		null,		null,

	";",		":",		"?",		"{",		"}",
	"(",		")",		"[",		"]",		"throws",
	"error",	"comment",	"type",		"length",	"inline-return",
	"inline-method","inline-new"
    };

inline ostream& printMod(ostream& o, int m) {
  if(m & ACC_PUBLIC)    o << "public ";
  if(m & ACC_PRIVATE)   o << "private ";
  if(m & ACC_PROTECTED) o << "protected ";
  if(m & ACC_STATIC   ) o << "static ";
  if(m & ACC_FINAL    ) o << "final ";
  if(m & ACC_SUPER    ) o << "super ";
  if(m & ACC_VOLATILE ) o << "volatile ";
  if(m & ACC_TRANSIENT) o << "transient ";
  if(m & ACC_NATIVE   ) o << "native ";
  if(m & ACC_INTERFACE) o << "interface ";
  if(m & ACC_ABSTRACT ) o << "abstract ";
  if(m & ACC_SYNCHRONIZED) o << "sync ";
  return o;
}

const int CONSTANT_Utf8        = 1;
const int CONSTANT_Unicode     = 2;
const int CONSTANT_Integer     = 3;
const int CONSTANT_Float       = 4;
const int CONSTANT_Long        = 5;
const int CONSTANT_Double      = 6;
const int CONSTANT_Class       = 7;
const int CONSTANT_String      = 8;
const int CONSTANT_Fieldref    = 9;
const int CONSTANT_Methodref   = 10;
const int CONSTANT_IMethodref  = 11;
const int CONSTANT_NameAndType = 12;

char* cpool_type_name_array[]={
  "bad","utf","uni","int","flo","lng","dbl",
  "cla","str","frf","mrf","irf","nat"
};
inline char* cpool_type(int i) {
  if(i > 0 && i <= 12) return cpool_type_name_array[i];
  else {
    static char a[4] = "000";
    a[0] = '0'+i/100;
    a[1] = '0'+(i/10)%10;
    a[2] = '0'+i%10;
    return a;
  }
}

const int TC_BOOL      = 0;
const int TC_BYTE      = 1;
const int TC_CHAR      = 2;
const int TC_SHORT     = 3;
const int TC_INT       = 4;
const int TC_LONG      = 5;
const int TC_FLOAT     = 6;
const int TC_DOUBLE    = 7;
const int TC_NULL      = 8;
const int TC_ARRAY     = 9;
const int TC_CLASS     = 10;
const int TC_VOID      = 11;
const int TC_METHOD    = 12;
const int TC_ERROR     = 13;

extern const char* typeNames[];

inline void printType(ostream& o, int typ) {
  if(typ < 0 || typ > TC_ERROR)
    o << "BAD_TYPE_" << typ;
  else
    o << typeNames[typ];
}

extern const char* opNames[];

const int TM_NULL      = 1 << TC_NULL;
const int TM_VOID      = 1 << TC_VOID;
const int TM_BOOL      = 1 << TC_BOOL;
const int TM_BYTE      = 1 << TC_BYTE;
const int TM_CHAR      = 1 << TC_CHAR;
const int TM_SHORT     = 1 << TC_SHORT;
const int TM_INT       = 1 << TC_INT;
const int TM_LONG      = 1 << TC_LONG;
const int TM_FLOAT     = 1 << TC_FLOAT;
const int TM_DOUBLE    = 1 << TC_DOUBLE;
const int TM_ARRAY     = 1 << TC_ARRAY;
const int TM_CLASS     = 1 << TC_CLASS;
const int TM_METHOD    = 1 << TC_METHOD;
const int TM_ERROR     = 1 << TC_ERROR;

const int TM_INT32     = TM_BYTE | TM_SHORT | TM_CHAR | TM_INT;
const int TM_NUM32     = TM_INT32 | TM_FLOAT;
const int TM_NUM64     = TM_LONG | TM_DOUBLE;
const int TM_INTEGER   = TM_INT32 | TM_LONG;
const int TM_REAL      = TM_FLOAT | TM_DOUBLE;
const int TM_NUMBER    = TM_INTEGER | TM_REAL;
const int TM_REFERENCE = TM_ARRAY | TM_CLASS | TM_NULL;

const int CS_UNDEFINED        = 0;
const int CS_UNDECIDED        = 1;
const int CS_BINARY           = 2;
const int CS_SOURCE           = 3;
const int CS_PARSED           = 4;
const int CS_COMPILED         = 5;
const int CS_NOTFOUND         = 6;

/*
  CONSTANT_Fieldref_info { // Methodref / Interfaceref
    u1 tag;
		u2 class_index;
    u2 name_and_type_index;
  }
	CONSTANT_NameAndType {
    u1 tag;
    u2 name_index;       // simple identifier -> CONSTANT_Utf8_info
    u2 descriptor_index; // -> CONSTANT_Utf8_info
  }
  CONSTANT_String_info {
    u1 tag;
		u2 string_index; // -> CONSTANT_Utf8_info
  }
  CONSTANT_Utf8_info {
    u1 tag;
    u2 length;
    u1 bytes[];
  }
-----------
	field_info = method_info {
    u2 access_flags;
    u2 name_index;        // -> Utf8_info
		u2 descriptor_index;  // -> Ut8
    u2 attributes_count;
    attribute_info attributes[ attributes_count ];
  }
-----------
*/
//

const int M_PUBLIC           = ACC_PUBLIC;
const int M_PRIVATE          = ACC_PRIVATE;
const int M_PROTECTED        = ACC_PROTECTED;
const int M_STATIC           = ACC_STATIC;
const int M_TRANSIENT        = ACC_TRANSIENT;
const int M_SYNCHRONIZED     = ACC_SYNCHRONIZED;
const int M_ABSTRACT         = ACC_ABSTRACT;
const int M_NATIVE           = ACC_NATIVE;
const int M_FINAL            = ACC_FINAL;
const int M_VOLATILE         = ACC_VOLATILE;
const int M_INTERFACE        = ACC_INTERFACE;

//    Modifier masks

const int MM_CLASS  = M_PUBLIC | M_INTERFACE | M_FINAL | M_ABSTRACT;
const int MM_FIELD  = M_PUBLIC | M_PRIVATE | M_PROTECTED |
		      M_FINAL  | M_STATIC | M_TRANSIENT | M_VOLATILE;
const int MM_METHOD = M_PUBLIC | M_PRIVATE | M_PROTECTED |
		      M_FINAL | M_STATIC | M_SYNCHRONIZED |
		      M_ABSTRACT | M_NATIVE;

enum {
  COMMA              = 0,
  ASSIGN             = 1,

  ASGADD             = 2,
  ASGSUB             = 3,
  ASGMUL             = 4,
  ASGDIV             = 5,
  ASGREM             = 6,

  ASGLSHIFT          = 7,
  ASGRSHIFT          = 8,
  ASGURSHIFT         = 9,
  ASGBITAND          = 10,
  ASGBITOR           = 11,
  ASGBITXOR          = 12,

  COND               = 13,
  OR                 = 14,
  AND                = 15,
  BITAND             = 16,
  BITOR              = 17,
  BITXOR             = 18,
  NE                 = 19,
  EQ                 = 20,

  LT                 = 21,
  GE                 = 22,
  GT                 = 23,
  LE                 = 24,

  INSTANCEOF         = 25,
  LSHIFT             = 26,
  RSHIFT             = 27,
  URSHIFT            = 28,
  ADD                = 29,
  SUB                = 30,
  MUL                = 31,
  DIV                = 32,
  REM                = 33,
  CAST               = 34,           // (x)y
  POS                = 35,           // +x
  NEG                = 36,           // -x
  NOT                = 37,
  BITNOT             = 38,
  PREINC             = 39,           // ++x
  PREDEC             = 40,           // --x
  NEWARRAY           = 41,
  NEWINSTANCE        = 42,
  NEWFROMNAME        = 43,
  POSTINC            = 44,           // x++
  POSTDEC            = 45,           // x--
  FIELD              = 46,
  METHOD             = 47,           // x(y)
  ARRAYACCESS        = 48,           // x[y]
  NEW                = 49,
  INC                = 50,
  DEC                = 51,

  CONVERT            = 55,           // implicit conversion
  EXPR               = 56,           // (x)
  ARRAY              = 57,           // {x, y, ...}
  GOTO               = 58,

  IDENT              = 60,
  BOOLEANVAL         = 61,
  BYTEVAL            = 62,
  CHARVAL            = 63,
  SHORTVAL           = 64,
  INTVAL             = 65,
  LONGVAL            = 66,
  FLOATVAL           = 67,
  DOUBLEVAL          = 68,
  STRINGVAL          = 69,

  BYTE		= 70,
  CHAR		= 71,
  SHORT		= 72,
  INT		= 73,
  LONG		= 74,
  FLOAT		= 75,
  DOUBLE		= 76,
  VOID		= 77,
  BOOLEAN		= 78,

//		 * Expression keywords
  TRUE		= 80,
  FALSE		= 81,
  THIS		= 82,
  SUPER		= 83,
  NULL_E		= 84,

//		 * Statement keywords
  IF			= 90,
  ELSE		= 91,
  FOR		= 92,
  WHILE		= 93,
  DO			= 94,
  SWITCH		= 95,
  CASE		= 96,
  DEFAULT		= 97,
  BREAK		= 98,
  CONTINUE		= 99,
  RETURN		= 100,
  TRY		= 101,
  CATCH		= 102,
  FINALLY		= 103,
  THROW		= 104,
  STAT		= 105,
  EXPRESSION		= 106,
  DECLARATION	= 107,
  VARDECLARATION	= 108,

  SYNCHRONIZED	= 126,

  ERROR		= 145,
  TYPE		= 147,
  LENGTH		= 148,
};

/*  Class File {
 *    u4 magic;
 *    u2 minor_version;
 *    u2 major_version;
 *    u2 constant_pool_count;
 *    cp_info constant_pool[ constant_pool_count - 1 ];
 *    u2 access_flags;
 *    u2 this_class;
 *    u2 super_class;
 *    u2 interfaces_count;
 *    u2 interfaces[ interfaces_count ];
 *    u2 fields_count;
 *    field_info fields;
 *    u2 methods_count;   
 *    method_info methods[ methods_count ];
 *    u2 attributes_count;
 *    attribute_info attributes[ attributes_count ];
 *  }
 */


