typedef unsigned char  u1;
typedef unsigned short u2;
typedef unsigned int   u4;

union bipart_hack { 
  long long asLong; 
  double    asDouble;
  struct {u4 hi, lo;} part; 
};

struct JavaDouble {
  u4 hi;
  u4 lo;

  JavaDouble(double val) {
    bipart_hack hack;
    hack.asDouble = val;
    hi = hack.part.hi;
    lo = hack.part.lo;
  }

  operator double() const {
    bipart_hack hack;
    hack.part.hi = hi;
    hack.part.lo = lo;
    return hack.asDouble;
  }
};

struct JavaLong {
  u4 hi;
  u4 lo;

  JavaLong(long long val) {
    bipart_hack hack;
    hack.asLong = val;
    hi = hack.part.hi;
    lo = hack.part.lo;
  }

  operator long long() const {
    bipart_hack hack;
    hack.part.hi = hi;
    hack.part.lo = lo;
    return hack.asLong;
  }
};

class ConstantPool {
  ConstantPool(ConstantPool&):buf(buf1){assert(false);}
protected:
  union Item {
    u1  asU1;
    u2  asU2;
    u4  asU4;
    struct {u2 class_ndx; u2 name_and_type_ndx;} method, field;
    struct {u2 name_utf; u2 type_utf;} name_and_type;
    struct {u2 name_utf;} clazz;
    char* asString;
    float asFloat;
    int asInteger;
  };

  int constant_pool_count;
  int capacity;
  Item * item;
  char * type;
  Buf& buf;

  u2 longM1;

public:
  int Code;          // favorite attributes
  int ConstantValue; //

  void reserve(int size) {
    if(capacity < size) {
      ASSERT(false, Sorry_guys, size);
      Item * items = new Item[size + (3+size)/4];
      char * types = (char*)(items + size);

      if(item) {
	memcpy(items, item, sizeof(item) * constant_pool_count); 
	memcpy(types, type,            1 * constant_pool_count); 
        delete[] item; // delete should not fail
      }
      item = items;
      type = types;
      capacity = size;
    }
  }
  
  /*
  explicit ConstantPool(int size = 0):Code(0),ConstantValue(0),capacity(0),
    constant_pool_count(1),type(0),item(0)
  {
    assert(false);
    assert(sizeof(Item) == 4);
    if(size > 0) reserve(size);
  }
  */

  explicit ConstantPool(Buf& b):Code(0),ConstantValue(0),capacity(b.len/8),
    constant_pool_count(1),type(0),item((Item*)b.mem()),buf(b),longM1(0)
  {
    type = (char*)(item + capacity);
    assert(sizeof(Item) == 4);
    //    if(size > 0) reserve(size);
  }

  //  ~ConstantPool() {  delete[] item;  }
  ~ConstantPool() {  buf.release();}
  
  void wantMore(int howMuch) {
    if(constant_pool_count + howMuch > capacity) reserve(2 * capacity);
  }

  u2 addClass(u2 utf) {
    wantMore(1); u2 current = constant_pool_count++;
    type[current] = CONSTANT_Class;
    item[current].clazz.name_utf = utf;
    return current;
  }

  u2 addFieldRef(int flavor, u2 clazz, u2 name_type) {
    wantMore(1); u2 current = constant_pool_count++;
    type[current] = flavor;
    item[current].method.class_ndx = clazz;
    item[current].method.name_and_type_ndx = name_type;
    return current;
  }

  u2 addNameAndType(u2 name_utf, u2 type_utf) {
    wantMore(1); u2 current = constant_pool_count++;
    type[current] = CONSTANT_NameAndType;
    item[current].name_and_type.name_utf = name_utf;
    item[current].name_and_type.type_utf = type_utf;
    return current;
  }

  u2 addString(u2 utf) {
    wantMore(1); u2 current = constant_pool_count++;
    type[current] = CONSTANT_String;
    item[current].asU2 = utf;
    return current;
  }

  u2 addTwoWords(int flavor, u4 hi, u4 lo) {
    wantMore(2); u2 current = constant_pool_count;
    constant_pool_count += 2;
    type[current+0] = flavor;  type[current+1] = 0;
    item[current+0].asU4 = hi; item[current+1].asU4 = lo;
    return current;
  }

  u2 addDouble(u4 hi, u4 lo) { return addTwoWords(CONSTANT_Double, hi, lo);  }

  u2 addDouble(double value) { 
    union hack { double asDouble; struct {u4 hi, lo;} b; } a;
    a.asDouble = value;
    assert(sizeof(double)==8 && sizeof(union hack)==8);
    return addTwoWords(CONSTANT_Double, a.b.hi, a.b.lo);  
  }

  u2 addLong(u4 hi, u4 lo) { return addTwoWords(CONSTANT_Long, hi, lo);  }
  u2 addLong(long value)   { return addLong(0,value); }

  u2 getLongM1() {
    if(!longM1) longM1 = addLong(~0,~0);
    return longM1;
  }

  u2 addInteger(int value) {
    wantMore(1); u2 current = constant_pool_count++;
    type[current] = CONSTANT_Integer;
    item[current].asInteger = value;
    return current;
  }

  u2 addFloat(float value) {
    wantMore(1); u2 current = constant_pool_count++;
    type[current] = CONSTANT_Float;
    item[current].asFloat = value;
    return current;
  }

  u2 addUtf(char * str) {
    wantMore(1); u2 current = constant_pool_count++;
    type[current] = CONSTANT_Utf8;
    item[current].asString = str;
    return current;
  }

  void prepareAttributeNames() {
    for(int i = 1; i < constant_pool_count; ++i)
      if(type[i] == CONSTANT_Utf8) {
        char* str = item[i].asString;
        if(!Code && strcmp(str,"Code") == 0)Code = i;
        if(!ConstantValue && strcmp(str,"ConstantValue") == 0)ConstantValue = i;
      }
  }

  char* getUtf(int i) {
    assert(type[i] == CONSTANT_Utf8);
    return item[i].asString;
  }

  char* getClassName(int i) {
    assert(type[i] == CONSTANT_Class);
    return getUtf(item[i].clazz.name_utf);
  }

  void printClass(ostream& o, int i) {
    assert(type[i] == CONSTANT_Class);
    o << "class_";
    printUtf(o, item[i].clazz.name_utf);
  }

  void printNameAndType(ostream& o, int i) {
    assert(type[i] == CONSTANT_NameAndType);
    printUtf(o, item[i].name_and_type.name_utf); o << " : ";
    printUtf(o, item[i].name_and_type.type_utf);
  }

  void printRef(ostream& cout, int ref_type, int i) {
    assert(type[i] == ref_type);
    printClass(cout, item[i].method.class_ndx);
    cout << ".";
    printNameAndType(cout, item[i].method.name_and_type_ndx);
  }

  void printRef(ostream& cout, int i) {
    switch(type[i]) {
      case CONSTANT_Fieldref:   cout << "F "; break;
      case CONSTANT_Methodref:  cout << "M "; break;
      case CONSTANT_IMethodref: cout << "I "; break;
      default: ASSERT(false, ER_NOT_REF, type[i]);
    }
    printClass(cout, item[i].method.class_ndx);
    cout << ".";
    printNameAndType(cout, item[i].method.name_and_type_ndx);
  }

  void printConst(ostream& cout, int i) {
    switch( type[i] ) {
    case CONSTANT_String:
      printUtf(cout, item[i].asU2); break;
      
    case CONSTANT_Integer:
      cout << (signed) item[i].asU4; break;
      
    case CONSTANT_Float:
      cout << item[i].asFloat; break;
      
    case CONSTANT_Long: cout << "(LONG) " << item[i].asU4 << ' '
			     << item[i+1].asU4; break;

    case CONSTANT_Double:   cout << "(DOUBLE)";  break;
    
    default:  cout << "BAD_CONSTANT_" << type[i];
    }
  }

  void printUtf(ostream& cout, int i) {
    cout << getUtf(i);
  }

  bool compareUtfWith(int i, const char* s) {
    return strcmp(s, getUtf(i)) == 0;
  }

  void print(ostream& cout) {
    cout << " size: " << constant_pool_count << " capacity: " << capacity << endl;
    for(int i = 1; i < constant_pool_count; ++i) {
      cout << setw(3) << i << ' ' << cpool_type(type[i]) << ": ";
      switch( type[i] ) {
        case CONSTANT_Class:    printClass(cout, i); break;

        case CONSTANT_Fieldref:
        case CONSTANT_Methodref:
        case CONSTANT_IMethodref:printRef(cout, type[i], i); break;

        case CONSTANT_String:
        case CONSTANT_Integer:
        case CONSTANT_Float: printConst(cout, i); break;
        case CONSTANT_Long:
        case CONSTANT_Double: printConst(cout, i); ++i; break;

        case CONSTANT_NameAndType: printNameAndType(cout, i); break;

        case CONSTANT_Utf8:     printUtf(cout, i); break;
        default:
           cout << "Unknown type: " << (int)type[i] << ' ';
      }
      cout << endl;
    }
  }

  void write(ostream& o, void* buf, int size) {
    o.write(buf,size);
  }
  void writeUtf(ostream& o, int i) {
    char* s = item[i].asString;
    u2 len = u2(*(s-1)) + 256 * u2(*(s-2));
    write(o, s-2, len+2);
  }
  void write2(ostream& o, u2 value) {
    o << (unsigned char)(value >> 8) << (unsigned char)(value & 255);
  }
  void write4(ostream& o, u4 value) {
    write2(o, value >> 16); write2(o, value & 65535u);
  }

  void write(ostream& o) {
    write2(o, constant_pool_count);
    for(int i = 1; i < constant_pool_count; ++i) {
      o << type[i];
      switch( type[i] ) {
        case CONSTANT_Class:    write2(o,item[i].clazz.name_utf); break;

        case CONSTANT_Fieldref: 
        case CONSTANT_Methodref:
        case CONSTANT_IMethodref: write2(o,item[i].field.class_ndx);
	                          write2(o,item[i].field.name_and_type_ndx); break;

        case CONSTANT_String:  write2(o,item[i].asU2); break;
        case CONSTANT_Integer: write4(o,item[i].asU4); break;

        case CONSTANT_Float:   write(o, &item[i].asFloat, 4); break;
        case CONSTANT_Long:
        case CONSTANT_Double: 
	                       write(o, &item[i].asU4, 4); write(o, &item[++i].asU4, 4); break;

        case CONSTANT_NameAndType: write2(o,item[i].name_and_type.name_utf);
	                           write2(o,item[i].name_and_type.type_utf); break;

        case CONSTANT_Utf8:    
				  writeUtf(o,i); break;
        default:
           cerr << "Unknown type: " << (int)type[i] << ' ';
      }
    }
  }
};

ConstantPool cpool(buf1);
