/************************************************
  Team: Jon Gregory, Gor Nishanov
  Project #3 
    for RPI course 66-648 Compiler Design Theory

  expr.h: 
    Expression classes definitions.
    
    to do:  ArrayInitializer

    *****/


class Node : public QuickNew {
  Node(Node&){assert(false);}
public:
  int op;
public:
  Node(int Op):op(Op) {}
  int getOp() {	return op; }
  virtual void print(ostream& o) const {
    o << "Nodes is not supposed to be here !!!!" << endl;
    assert(false);
  }

  void nim(const char* s) { 
    cerr << s << " is FORBIDDEN (and/or not implemented)" << op << " " 
	 << opNames[op] << ' ' << this << endl;
    abort();
  }

  static void safePrint(ostream& o, const Node* e) {
    if(e == 0)
      o << "<null>";
    else o << *e;
  }
  friend ostream& operator << (ostream& o, const Node& n) {
    n.print(o);
    return o;
  }
  friend ostream& operator << (ostream& o, const Node* n) {
    Node::safePrint(o, n);
    return o;
  }
  void printOpName(ostream& o) const { o << ('(' + opNames[op] + ' '); }
};


class Expression : public Node {
public:
  mutable Type* type;
public:
  Type* getType() { return type; }
  Expression(int op, Type* typ):Node(op),type(typ),nextInOwner(0) {}

  virtual bool isConstant() {	return false;	}
  virtual Identifier toIdent() const {return Identifier::crazy;}

  // The following is for use only in VariableId //
  virtual void increaseDim() 
    { ASSERT(false, ER_NOT_AN_IDENTIFIER, toIdent() ); }
  virtual int getDims() {
    ASSERT(false, ER_NOT_AN_IDENTIFIER, toIdent() );
    return 0;
  }

  virtual Context* ctxValue(bool rval = true) {nim("ctxValue"); return 0; }
  virtual int codeLValue()  { nim("codeLValue"); return 0; }
  virtual void codeValue()  { nim("codeValue"); }
  virtual void codeStore()  { nim("codeStore"); }
  virtual void codeLoad()   { nim("codeLoad"); }

  void boolToValue() {
    assert(type == Type::Bool);
    int forward   = code.branch(op_goto, 0); // to patch
    int falseAddr = code.currentAddress();
      code.add(op_iconst_0);
    int end       = code.branch(op_goto, 0); // to patch
    code.patch_branch(forward);
    codeBranchIfFalse(falseAddr);
    code.add(op_iconst_1);
    code.patch_branch(end);
  }

  virtual void codeVoid() {
    codeValue();    
    switch (type->typeCode) {
      case TC_VOID:   case TC_ERROR: break;
      case TC_DOUBLE: case TC_LONG: code.add(op_pop2); break;
      default: code.add(op_pop);
    }
  }

  void codeDup(unsigned items, unsigned depth) {
    ASSERT(depth < 3, ER_BAD_DUP_DEPTH, depth);
    ASSERT(items < 3, ER_BAD_DUP_ITEMS, items);
    if(items == 0) return;
    code.add(op_dup + --items * 3 + depth);
  }

  virtual bool notIntZero() { return true; }

  virtual bool isLocal() { return false; }
  virtual bool isString() { return false; } 
  virtual Expression* getLeft() { assert(false); return 0; }
  virtual Expression* getRight() { assert(false); return 0; }
  virtual Expression* getDeclaringIdent() { 
    ASSERT(false, ER_NOT_AN_IDENTIFIER, toIdent() );
    return 0;
  }
  virtual void importClass() 
    { ASSERT(false, ER_CANNOT_IMPORT_THIS, toIdent() ); }

  virtual MethodCtx * prepareMethodCall() {
    ASSERT(false, ER_SHOULD_NOT_DO_THIS, toIdent());
    return 0;
  }  
  virtual PackageCtx * resolvePackageName() {
    ASSERT(false, ER_NOT_A_PACKAGE, toIdent());
    return 0;
  }
  virtual ClassCtx* resolveClassName() {
    ASSERT(false, ER_NOT_A_CLASS, toIdent());
    return 0;
  }
  virtual void codeBranchIfFalse( int addr ) {
    codeValue();
    ASSERT(type == Type::Bool, ER_IT_IS_NOT_A_C, this);
    code.branch(op_ifeq, addr);
  }
  virtual void codeBranchIfTrue( int addr ) {
    codeValue();
    ASSERT(type == Type::Bool, ER_IT_IS_NOT_A_C, this);
    code.branch(op_ifne, addr);
  }
  virtual bool isSuper() { return false; }
  
  virtual void print(ostream& o) const { o << opNames[op];	}

  static void codeConversion(Type* f, Type* t) {
    int from = f->typeCode;
    int to   = t->typeCode;
    
    switch (to) {
    case TC_BOOL: 
      if (from != TC_BOOL) break;
      return;
    case TC_BYTE:
      if (from != TC_BYTE) {
	codeConversion(f, Type::Int);
	code.add(op_int2byte);
      }
      return;
    case TC_CHAR:
      if (from != TC_CHAR) {
	codeConversion(f, Type::Int);
	code.add(op_int2char);
      }
      return;
    case TC_SHORT:
      if (from != TC_SHORT) {
	codeConversion(f, Type::Int);
	code.add(op_int2short);
      }
      return;
    case TC_INT:
      switch (from) {
      case TC_BYTE:
      case TC_CHAR:
      case TC_SHORT:
      case TC_INT:
	return;
      case TC_LONG:
	code.add(op_l2i);
	return;
      case TC_FLOAT:
	code.add(op_f2i);
	return;
      case TC_DOUBLE:
	code.add(op_d2i);
	return;
      }
      break;
    case TC_LONG:
      switch (from) {
      case TC_BYTE:
      case TC_CHAR:
      case TC_SHORT:
      case TC_INT:
	code.add(op_i2l);
	return;
      case TC_LONG:
	return;
      case TC_FLOAT:
	code.add(op_f2l);
	return;
      case TC_DOUBLE:
	code.add(op_d2l);
	return;
      }
      break;
    case TC_FLOAT:
      switch (from) {
      case TC_BYTE:
      case TC_CHAR:
      case TC_SHORT:
      case TC_INT:
	code.add(op_i2f);
	return;
      case TC_LONG:
	code.add(op_l2f);
	return;
      case TC_FLOAT:
	return;
      case TC_DOUBLE:
	code.add(op_d2f);
	return;
      }
      break;
    case TC_DOUBLE:
      switch (from) {
      case TC_BYTE:
      case TC_CHAR:
      case TC_SHORT:
      case TC_INT:
	code.add(op_i2d);
	return;
      case TC_LONG:
	code.add(op_l2d);
	return;
      case TC_FLOAT:
	code.add(op_f2d);
	return;
      case TC_DOUBLE:
	return;
      }
      break;
      
    case TC_CLASS:
      switch (from) {
      case TC_NULL: case TC_CLASS: case TC_ARRAY: return;
      }
      break;
    case TC_ARRAY:
      switch (from) {
      case TC_NULL: case TC_CLASS: case TC_ARRAY: return;
      }
      break;
    }
    ASSERT2(false, ER_INVALID_CAST, f, t);
  }
  
  Expression * nextInOwner;
};

typedef List<Expression> ExprList;
ExprList empty_exprs;

class ConstantExpression : public Expression {
public:
  ConstantExpression(int op, Type* type):Expression(op, type){}

  bool isConstant() {	return true;}
};


class UnaryExpression: public Expression {
protected:
  Expression* right;
public:
  UnaryExpression(int op, Type* type, Expression* r):
    Expression(op, type),right(r) {}
  
  virtual void print(ostream& o) const {
    o << '(' << opNames[op] << ' ' << right << ')';
  }
};

class NaryExpression : public UnaryExpression {
protected:
  Expression * args; //[]
  int args_count;
public:
  NaryExpression(int op, Type *type, Expression *right, ExprList* arg) :
    UnaryExpression(op, type, right),args(arg->first),args_count(arg->count) {}
  
  virtual void print(ostream& o) const {
    o << '(' << opNames[op];
    if (right != NULL) {
      o << " ";
      right->print(o);
    }
    Expression *e = args;
    if(e == 0) o << "<null>";
    while(e != 0) {
      o << ' ' << e;
      e = e->nextInOwner;
    }
    o << ')';
  }
};

class ArrayInitializerExpression : public NaryExpression {
public:
  ArrayInitializerExpression(ExprList* args):
    NaryExpression(ARRAY, Type::Error, NULL, args){}
};

class BinaryExpression : public UnaryExpression {
protected:
  Expression* left;
public:
  BinaryExpression(int op, Type* type, Expression* Left, Expression* right):
    UnaryExpression(op, type, right),left(Left){}
  
  virtual void print(ostream& o) const {
    o << '(' << opNames[op] << ' ';
    safePrint(o, left);
    o << ' ';
    safePrint(o, right);
    o << ')';
  }
};

class BinaryAssignExpression : public BinaryExpression {
public:
  BinaryAssignExpression(int op, Expression *left, Expression *right):
    BinaryExpression(op, left->type, left, right){}
};

class AssignExpression : public BinaryAssignExpression {
public:
  AssignExpression(Expression* left, Expression* right):
    BinaryAssignExpression(ASSIGN, left, right){}
  virtual Expression* getDeclaringIdent() { return left; }

  virtual void codeValue() {
    int depth = left->codeLValue();
    right->codeValue();
    if(right->type->inMask(TM_INT32)) {
      codeConversion(right->type, left->type);
      type = left->type;
    } else {
      type = right->type;
    }
    ASSERT(type == left->type, ER_EXPLICIT_CAST_NEEDED, this);
    codeDup(type->stackSize(), depth);
    left->codeStore();
  }
  virtual void codeVoid() {
    //    cerr << "AssignExpression::codeVoid" << endl;
    left->codeLValue();
    right->codeValue();
    if(right->type->inMask(TM_INT32)) {
      codeConversion(right->type, left->type);
      type = left->type;
    } else {
      type = right->type;
    }
    ASSERT(type == left->type, ER_EXPLICIT_CAST_NEEDED, this);
    left->codeStore();
  }
};

class ArrayAccessExpression : public UnaryExpression {
  Expression* index;
public:
  ArrayAccessExpression(Expression* right, Expression* ndx):
    UnaryExpression(ARRAYACCESS, Type::Error, right),index(ndx){}

  void codeLoad()  { code.add(op_iaload  + type->aoffset() );  }
  void codeStore() { code.add(op_iastore + type->aoffset() );  }

  int codeLValue() {
    right->codeValue();
    index->codeValue();
    type = right->type->getElementType();
    return 2;
  }

  void codeValue() {
    codeLValue();
    codeLoad();
    if(type->inMask(TM_INT32)) type = Type::Int;
  }
  
  virtual void print(ostream& o) const {
    o << '(' << opNames[op] << ' ';
    safePrint(o, right);
    o << ' ' << index << ')';
  }
};

class ArrayGenerateExpression : public Expression {
  Expression* indices;
  int count;
public:
  ArrayGenerateExpression(Type* type, List<Expression>& ndx):
    Expression(ARRAYACCESS, type),indices(ndx.first),count(ndx.count){}

  virtual void codeValue() {
    assert(count);
    for(Expression* i = indices; i; i = i->nextInOwner) i->codeValue();
    if(count == 1) {
      Type* elem = type->getElementType();
      if(elem->inMask(TM_NUMBER)) 
	code.add(op_newarray).arrayCode(elem);
      else
	code.add(op_anewarray).type(elem);
    } else {
      code.add(op_multianewarray, count, 1).class_type(type);
      code.byte(count);
    }
  }
  
  virtual void print(ostream& o) const {
    o << "(new " << type;
    for(Expression* i = indices; i; i = i->nextInOwner) {
      o << ' ' << i;
    }
    o << ']';
  }
};

class BinaryCompareExpression : public BinaryExpression {
public:
  BinaryCompareExpression(int op, Expression *left, Expression *right):
    BinaryExpression(op, Type::Bool, left, right){ }

  // F = 0 => if true, F = 1 if false  
  virtual void codeValue() { boolToValue(); }

  void ifFalseOrTrue( int addr, int F ) {
    left->codeValue();
    if(right->notIntZero()) right->codeValue();

    ASSERT(left->type == right->type, ER_SORRY_NO_PROMOTIONS, this);

    if(left->type->inMask(TM_INT32)) {
      int opc = (right->notIntZero())?op_if_icmplt:op_iflt;
      opc += (op - LT)^F;
      code.branch(opc, addr);
    } else {
      switch(left->type->typeCode) {
        case TC_LONG: code.add(op_lcmp); break;
        case TC_FLOAT: code.add(op_fcmpl); break;
        case TC_DOUBLE: code.add(op_dcmpl); break;
      default:
	ASSERT(false, ER_CAN_NOT_COMPARE, left->type);
      }
      code.branch( (op-LT)^F + op_iflt , addr);
    }
  }
  virtual void codeBranchIfFalse( int addr ) { ifFalseOrTrue(addr, 1); }
  virtual void codeBranchIfTrue( int addr ) { ifFalseOrTrue(addr, 0); }
};

class BinaryBitExpression : public BinaryExpression {
public:
  BinaryBitExpression(int op, Expression *left, Expression *right):
    BinaryExpression(op, left->type, left, right) {}

  virtual void codeValue() {
    left->codeValue();
    right->codeValue();
    ASSERT(left->type == right->type, ER_SORRY_NO_PROMOTIONS, this);
    type = left->type;
    int L = (type == Type::Long)?1:0;
    code.add(op_iand + L + (op - BITAND)*2);
  }

};

class BinaryEqualityExpression : public BinaryExpression {
public:
  BinaryEqualityExpression(int op, Expression *left, Expression *right):
    BinaryExpression(op, Type::Bool, left, right) {}

  virtual void codeValue() { boolToValue(); }

  void equals(int F, int trueAddr) { // F = 1 to negate // 
    if(op == NE) F = 1 - F;
    if(left->op == NULL_E) {
      right->codeValue();
      code.branch(op_ifnull + F, trueAddr);
      return;
    }
    if(right->op == NULL_E) {
      left->codeValue();
      code.branch(op_ifnull + F, trueAddr);
      return;
    }
    left->codeValue();
    right->codeValue();
    if(left->type->inMask(TM_REFERENCE)) {
      code.branch(op_if_acmpeq + F, trueAddr);
      return;
    }
    ASSERT(left->type == right->type, ER_SORRY_NO_PROMOTIONS, this);
    if(left->type->inMask(TM_INT32)) {
      code.branch(op_if_icmpeq + F, trueAddr);
    } else {
      switch(left->type->typeCode) {
        case TC_LONG: code.add(op_lcmp); break;
        case TC_FLOAT: code.add(op_fcmpl); break;
        case TC_DOUBLE: code.add(op_dcmpl); break;
      default:
	ASSERT(false, ER_CAN_NOT_COMPARE, left->type);
      }
      code.branch( op_ifeq + F , trueAddr);
    }
  }

  virtual void codeBranchIfFalse(int falseAddr) { equals(1, falseAddr); }
  virtual void codeBranchIfTrue(int trueAddr)   { equals(0, trueAddr); }
};

class BinaryLogicalExpression: public BinaryExpression {
public:
  BinaryLogicalExpression(int op, Expression *left, Expression *right):
    BinaryExpression(op, Type::Bool, left, right) {}

  virtual void codeValue() { boolToValue(); }

  virtual void codeBranchIfFalse( int falseAddr ) {
    if(op == OR) {
      int gotoAddr = code.branch(op_goto, 6); // skip the next goto
      int trueAddr = code.branch(op_goto, 0); // to patch later
      code.patch_branch(gotoAddr);
       left->codeBranchIfTrue(trueAddr);
      right->codeBranchIfFalse(falseAddr);
      code.patch_branch(trueAddr);
    } else {
       left->codeBranchIfFalse(falseAddr);
      right->codeBranchIfFalse(falseAddr);
    }
  }

  virtual void codeBranchIfTrue( int trueAddr ) {
    if(op == AND) {
      int  gotoAddr = code.branch(op_goto, 6); // skip the next goto
      int falseAddr = code.branch(op_goto, 0); // to patch later
      code.patch_branch(gotoAddr);
      left->codeBranchIfFalse(falseAddr);
      right->codeBranchIfTrue(trueAddr);
      code.patch_branch(falseAddr);
    } else {
       left->codeBranchIfTrue(trueAddr);
      right->codeBranchIfTrue(trueAddr);
    }
  }

};


class BoolExpression : public ConstantExpression {
  bool value;
public:
  BoolExpression(bool val) :
    ConstantExpression(BOOLEANVAL, Type::Bool),value(val){}

  virtual void codeValue() {
    code.add(value ? op_iconst_1 : op_iconst_0 );
  }
  
  virtual void print(ostream& o) const {
    o << ((value) ? "true" : "false");}
};

class IntegerExpression : public ConstantExpression {
public:
  int value;
  IntegerExpression(int op, Type* type, int val):
    ConstantExpression(op, type),value(val) {}

  virtual bool notIntZero() { return value != 0; }

  void codeValue() { code.load_int_const(value);  }
};

class ByteExpression : public IntegerExpression {
public:
  ByteExpression(int value):
    IntegerExpression(BYTEVAL, Type::Byte, value){}
  
  virtual void print(ostream& o) const {
    o << value << 'b'; }
};

class CastExpression : public UnaryExpression {
public:
  CastExpression(Type *type, Expression *right):
    UnaryExpression(CAST, type, right){}

  virtual void codeValue() {
    right->codeValue();
    codeConversion(right->type, type);
  }

  virtual void codeVoid() { right->codeVoid(); }
  
  virtual void print(ostream& o) const {
    o << '(' << opNames[op] << ' ' << type << ' ';
    safePrint(o, right); o << ')';
  }
};

class CharExpression : public IntegerExpression {
public:
  CharExpression(char value):
    IntegerExpression(CHARVAL, Type::Char, value) {}
  
  virtual void print(ostream& o) const {
    o << '\'' << (char)value << '\'';
  }
};

class CommaExpression : public BinaryExpression {
public:
  CommaExpression(Expression* left, Expression* right):
    BinaryExpression(COMMA,
		     (right != null) ? right->type : Type::Void, left, right) {}
  
  virtual void codeValue() {
    if(left)  left->codeVoid();
    if(right) right->codeValue();
  }

  virtual void codeVoid() {
    if(left)  left->codeVoid();
    if(right) right->codeVoid();
  }
};

class ConditionalExpression : public BinaryExpression {
  Expression* cond;
public:
  ConditionalExpression(Expression *c, Expression *left, Expression *right):
    BinaryExpression(COND, Type::Error, left, right), cond(c) {}

  virtual void codeValue() {
    int gotoAddr = code.branch(op_goto, 0);
    int falseAddr = code.currentAddress();
    right->codeValue();
    int  brkAddr = code.branch(op_goto, 0);
    code.patch_branch(gotoAddr);
    cond->codeBranchIfFalse(falseAddr);
    left->codeValue();
    code.patch_branch(brkAddr);
    ASSERT(left->type == right->type, ER_SORRY_NO_PROMOTIONS, this);
    type = left->type;
  }
  
  virtual void print(ostream& o) const {
    o << '(' << opNames[op] << ' ';
    safePrint(o, cond); o << ' ';
    safePrint(o, left); o << ' ';
    safePrint(o, right); o << ')';
  }
};

class DoubleExpression : public ConstantExpression {
  JavaDouble value;
public:
  DoubleExpression(double val): value(val),
    ConstantExpression(DOUBLEVAL, Type::Double){}

  virtual void codeValue() {
    if(value == 0.0) code.add(op_dconst_0);
    else if(value == 1.0) code.add(op_dconst_1);
    else {
      u2 ndx = cpool.addDouble(value);
      code.add(op_ldc2w); code.add2(ndx); 
    }
  }

  virtual void print(ostream& o) const {
    o << value << "D";}
};

class FieldExpression : public UnaryExpression {
  Identifier id;
  FieldCtx * field;
public:
  FieldExpression(Expression* right, Identifier Id):
    UnaryExpression(FIELD, Type::Error, right),id(Id),field(0) {}
  
  virtual void print(ostream& o) const {
    o << '(' << right << '.' << id  << ')';
  }
  
  virtual PackageCtx * resolvePackageName() {
    //    cerr << "Field::resolvePackageName(" << toIdent() << ')';
    Context * pkg = globalCtx.find(toIdent());
    return asPackageCtx( pkg );
  }

  virtual void importClass() {
    Context * pkg = right->resolvePackageName();
    globalCtx.importClass(pkg, id);
  }

  virtual ClassCtx * resolveClassName() { // We need to load that class
    Context * pkg = right->resolvePackageName();
    Context * clazz = pkg->safeLocalFind(id);
    ASSERT(clazz->type == Context::CLASS, ER_NOT_A_CLASS, toIdent());
    return (ClassCtx*) clazz;
  }
  
  virtual Context * ctxValue(bool rval) {
    Context * rightCtx = right->ctxValue(true);
    Context * ctx = 0;
    switch(rightCtx->type) {
    case Context::PACKAGE:
      return rightCtx->safeLocalFind(id);
    case Context::CLASS:
      // id should be a static object //
      ctx = rightCtx->safeLocalFind(id);
      ASSERT(ctx->isStatic(), ER_NON_STATIC_MEMBER, toIdent());
      if(rval && ctx->type == Context::FIELD) {
	field = (FieldCtx*) ctx;
	type = field->fieldType;
	code.add(op_getstatic, 0, type->stackSize()).ref(field->getRef());
      }
      return ctx;

    case Context::FIELD:
      type = ((FieldCtx*)(rightCtx))->fieldType;
      if( type->isArray() ) {
	ASSERT(Identifier::length == id, ER_NOT_A_CLASS, toIdent());
	type = Type::Int;
	code.add(op_arraylength);
	ASSERT(rval, ER_INT_CANNOT_BE_LVAL, this);
	return Context::errorCtx;
      }
      ctx = type->getClassCtx()->safeLocalFind(id);

      if(ctx->type == Context::FIELD) {
	field = (FieldCtx*)ctx;
	if(rval) field->issueLoad();
        type = ctx->getType();
      }
      return ctx;

    default:
      ASSERT(false, ER_SOMETHING_STRANGE, toIdent());
      return 0;
    }
  }
  
  virtual void codeValue() {
    ctxValue(true);
    if(type->inMask(TM_INT32)) type = Type::Int;
  }

  virtual int codeLValue() {
    ctxValue(false);
    return type->stackSize();
  }

  virtual void codeStore() {
    assert(field);
    field->issueStore(); 
  }

  virtual MethodCtx * prepareMethodCall() {
    //    cerr << "FieldExpression::prepareMethodCall()" << endl;
    return asMethodCtx(ctxValue(false));
  }
  
  virtual Identifier toIdent() const {
    char a[100];
    char* ptr = (right->toIdent()).copyTo(a);
    *ptr = '/'; id.copyTo(++ptr);
    return Identifier(a);
  }
};
class IncDecExpression;
class IdentifierExpression : public Expression {
  Identifier id;
  int dims;

  FieldCtx* field;

  friend IncDecExpression;
public:  
  IdentifierExpression(Identifier Id):field(0),
  Expression(IDENT, Type::Error),id(Id),dims(0){}

  virtual Expression* getDeclaringIdent() { return this; }

  void bind() {
    if(!field) {
      field = asFieldCtx(Context::current->find(id));
      type = field->fieldType;
    }
  }
  virtual bool isLocal() { 
    bind();
    assert(field);
    return field->isLocalVar(); 
  }
  virtual int codeLValue() { 
    bind();
    if(!field->isLocalVar() && !field->isStatic()) 
      { code.add(op_aload_0); return 1; } 
    return 0; 
  }

  virtual Context* ctxValue(bool rval) {
    Context* ctx = Context::current->find(id);
    if(ctx->type == Context::FIELD) {
      field = (FieldCtx*)ctx;
      if(!field->isLocalVar() && !field->isStatic()) code.add(op_aload_0);

      if(rval) field->issueLoad();

      type = field->fieldType;
    } else if(ctx->type == Context::CLASS) {
      type = ctx->getType();
    }
    return ctx;
  }

  virtual void codeLoad() {
    bind();
    field->issueLoad();
    if(type->inMask(TM_INT32)) type = Type::Int;
  }

  virtual void codeStore() {
    assert(field);
    field->issueStore();
  }

  void codeValue() {
    codeLValue();
    codeLoad();
  }

  virtual MethodCtx * prepareMethodCall() {
    MethodCtx* method = asMethodCtx( Context::current->find(id) );
    //    cerr << " method " << method->fullName 
    //	 << " is Static = " << method->isStatic() << endl;
    if(!(method->isStatic())) {
      ASSERT(!Context::current->isStatic(), ER_NON_STATIC_CALL, id);
      code.add(op_aload_0);
    }
    return method;
  }

  virtual ClassCtx * resolveClassName() { // We need to load that class
    return asClassCtx( globalCtx.mention(id) );
  }

  virtual PackageCtx * resolvePackageName() {
    return asPackageCtx( globalCtx.find(toIdent() ) );
  }

  // The following is for use only in VariableId //
  virtual void increaseDim() { ++dims; }
  virtual int getDims() { return dims; }

  virtual Identifier toIdent() const {return id;}
  virtual void print(ostream& o) const {o << id;}
};

inline bool childFather(Type* son, Type* parent) {
  if(son->typeCode != TC_CLASS) return false;
  return son->getClassCtx()->canReach(parent);
}

Prototype * MethodCtx::matchPrototype(Expression* exprs, int count) {
  for(Prototype* p = prototypes.first; p; p = p->nextInOwner) {
    TypeList* params = p->getType()->getArgumentTypes();
    //    cerr << "[MATCHING " << p->getType()->typeSig << ']' << exprs << ' ' << count 
    //         << ' ' << params->count << endl;
    if(params->count != count) continue;
    if(count == 0) return p;
    TypeListItem * t = params->first;
    Expression * e = exprs;
    while(e) {
      //      cerr << '{' << e->type << '}' << endl;
      if(e->type != t->type && !childFather(e->type, t->type)) break;
      e = e->nextInOwner;
      t = t->nextInOwner;
    }
    if(e == 0) return p;
  }
  srcPos(cerr) << "CANNOT_RESOLVE_OVERLOADING" << endl;
  cerr << " matching " << fullName << '(';
  for(Expression * e = exprs; e; e = e->nextInOwner) cerr << ' ' << e->type;
  cerr << ')' << endl;
  printPrototypes();
  abort();
  return 0;
}

class MethodExpression : public NaryExpression {

//  ClassDefinition clazz;
//  FieldDefinition field;

public:
  MethodExpression(Expression* right, ExprList* args):
    NaryExpression(METHOD, Type::Error, right, args){}

  virtual bool isSuper() { return right->isSuper(); }

  static Type* call(MethodCtx* method, Expression* args, int args_count) {
    int stack_before = 0;
    //    Context::current->find(Identifier("Font"));
    //if(stack_before == 0) return Type::Void;
    for(Expression* e = args; e; e = e->nextInOwner) { 
      e->codeValue(); stack_before += e->type->stackSize(); 
      if(e->type == Type::StringBuffer && 
         (method->fullName == Identifier::print ||
	  method->fullName == Identifier::println)) 
      {
	ClassCtx* ctx = Type::StringBuffer->getClassCtx();
        MethodCtx* method = asMethodCtx(ctx->safeLocalFind(Identifier::toString));
        MethodExpression::call(method, 0, 0);
	e->type = Type::String;
      }
    }
    Prototype* p = method -> matchPrototype(args, args_count);
    Type * type = p->getType()->getReturnType();
    int stack_after = type->stackSize();
    if(!method->isStatic()) ++stack_before;

    ASSERT(method->isStatic() == p->isStatic(), ER_PLEASE_BE_CONSISTENT, method->fullName);

    if(method->isInterface()) {
      code.add(op_invokeinterface, stack_before, stack_after);
      code.ref(p->getRef());
      code.byte(args_count); 
      code.byte(0); // reserved //
    } else {
      int op;

      if(method->fullName == Identifier::ctor) op = op_invokenonvirtual;
      else if(method->isStatic()) op = op_invokestatic;
      else op = op_invokevirtual;

      code.add(op, stack_before, stack_after).ref(p->getRef());
    }
    if(type->inMask(TM_INT32)) type = Type::Int;
    return type;
  }
  virtual void codeValue() {
    //    cerr << "MethodExpression::codeValue" << endl;
    MethodCtx* method = right->prepareMethodCall();
    type = call(method, args, args_count);
  }

  virtual void print(ostream& o) const {
    o << '(' << opNames[op] << ' ' << right;
    for(Expression * e = args; e; e = e->nextInOwner) o << ' ' << e;
    o << ')';
  }
};

class BinaryArithmeticExpression : public BinaryExpression {
public:
  BinaryArithmeticExpression(int op, Expression *left, Expression *right):
    BinaryExpression(op, left->type, left, right){}

  virtual bool isString() { return op == ADD && left->isString(); }
  virtual Expression* getLeft() {  return left; }
  virtual Expression* getRight() { return right; }

  void goLeft(MethodCtx* ctor, MethodCtx* append) {
    if(left->op == STRINGVAL) {
      code.add(op_new).class_type(Type::StringBuffer);
      code.add(op_dup);
      MethodExpression::call(ctor, left, 1);
    } else {
      ASSERT(left->op == ADD, ER_ONLY_ADDITIONS_ALLOWED, this);
      ((BinaryArithmeticExpression*)left)->goLeft(ctor, append );
    }
    MethodExpression::call(append, right, 1);
  }

  void recursiveCat() {
    //    cerr << "RECURSIVE CAT:" << endl;
    type = Type::StringBuffer;
    ClassCtx* ctx = type->getClassCtx();
    MethodCtx* append = asMethodCtx(ctx->safeLocalFind(Identifier::append));
    MethodCtx* ctor   = asMethodCtx(ctx->safeLocalFind(Identifier::ctor));
    goLeft(ctor, append);
  }

  void codeValue() {
    if(left->type == Type::String) {
      ASSERT(op == ADD, ER_ONLY_ADD_ALLOWED, this);
      recursiveCat();
      //      if(right->op == ADD) iterativeCat();
      //      else if(left->op == ADD) recursiveCat();
      //      else stringCat();
      return;
    } 
      
    left->codeValue();
    type = left->type;
    right->codeValue();
    type = left->type;
    ASSERT2(left->type == right->type, ER_SORRY_NOPROMOTIONS, left->type, right->type);
    assert(op >= ADD && op <= REM);
    code.add( op_iadd + 4*(op-ADD) + type->offset() );
  }
  
};

class FloatExpression : public ConstantExpression {
  float value;
  
public:
  FloatExpression(float val):
    ConstantExpression(FLOATVAL, Type::Float),value(val){}

  virtual void codeValue() {
    if(value == 0)      code.add(op_fconst_0);
    else if(value == 1) code.add(op_fconst_1);
    else if(value == 2) code.add(op_fconst_2);
    else {
      u2 ndx = cpool.addFloat(value); 
      if(ndx < 256) { code.add(op_ldc1); code.byte(ndx); }
      else {code.add(op_ldc2w); code.add2(ndx); }
    }
  }

  virtual void print(ostream& o) const {
    o << value << 'F' ;}
};

class IncDecExpression : public UnaryExpression {
  bool prefix;
  bool inc;
public:
  IncDecExpression(int op, Expression* right, bool Prefix):
    UnaryExpression(op, right->type, right),
    prefix(Prefix), inc(op == INC) 
  {}

  void codeIncDec(bool valNeeded) {
    if( right->isLocal() && right->type == Type::Int) {
      type = right->type;
      if (valNeeded && !prefix) right->codeLoad();
      int slot = ((IdentifierExpression*)right)->field->getSlot();
      code.add(op_iinc).iinc(slot, inc?1:-1);
      if (valNeeded && prefix) right->codeLoad();
      return;
    }
    int depth = right->codeLValue();
    codeDup(depth, 0);
    right->codeLoad();
    type = right->type;
    if (valNeeded && !prefix) {
      codeDup(type->stackSize(), depth);
    }
    if(type->inMask(TM_INT32)) {
      code.add(op_iconst_1);
      code.add(inc ? op_iadd : op_isub);
      if(type != Type::Int) code.add(op_int2byte + type->typeCode - TC_BYTE);
    } else {
      int tc = type->typeCode - TC_LONG;
      ASSERT2(tc >= 0 && tc <= 2, ER_INVALID_TYPE_FOR_INCDEC, type, this);
      static int ops[] = {op_lconst_1, op_fconst_1, op_dconst_1};
      code.add(ops[tc]);
      code.add(tc + (inc?op_ladd:op_lsub));
    }
    if (valNeeded && prefix) {
      codeDup(type->stackSize(), depth);
    }
    right->codeStore();
  }

  virtual void codeValue() { codeIncDec(true); } 
  virtual void codeVoid() { codeIncDec(false); }
};

class PreIncDecExpression : public IncDecExpression {
public:
  PreIncDecExpression(int op, Expression* right):
    IncDecExpression(op, right, true){}
};

class PostIncDecExpression : public IncDecExpression {
public:
  PostIncDecExpression(int op, Expression* right):
    IncDecExpression(op, right, false){}

  virtual void print(ostream& o) const {
    o << '(' << right << opNames[op] << ')';
  }  
};

class IntExpression : public  IntegerExpression {
public: IntExpression(int val):IntegerExpression(INTVAL, Type::Int, val){}
  virtual void print(ostream& o) const { o << value; }
};

// class LengthExpression : public UnaryExpression {
// public:
//   LengthExpression(Expression* right):
//     UnaryExpression(LENGTH, Type::Int, right){}
// };

class InstanceOfExpression : public BinaryExpression {
  Type * ofType;
public:
  InstanceOfExpression(Expression* left, Type* typ):
    BinaryExpression(INSTANCEOF, Type::Bool, left, 0),ofType(typ){}

  virtual void codeValue() {
    left->codeValue();
    code.add(op_instanceof).class_type(type);
  }
  
  virtual void print(ostream& o) const {
    printOpName(o);
    o << left << ' ';
    if (right && right->op == TYPE)
      o << right->type;
    else
      o << right;
    o << ')';
  }
};

class LongExpression : public ConstantExpression {
  JavaLong value;
public:
  LongExpression(long long val):
    ConstantExpression(LONGVAL, Type::Long), value(val){}
  
  virtual void codeValue() {
    if(value == 0) code.add(op_lconst_0);
    else if(value == 1) code.add(op_lconst_1);
    else {
      u2 ndx = cpool.addLong(value.hi, value.lo); 
      code.add(op_ldc2w); code.add2(ndx); 
    }
  }

  virtual void print(ostream& o) const {
    o << value << 'L';}
};

class BitNotExpression : public UnaryExpression {
public:
  BitNotExpression(Expression* right):
    UnaryExpression(BITNOT, right->type, right) {}

  virtual void codeValue() {
    right->codeValue();
    ASSERT(right->type->inMask(TM_INTEGER), ER_DEFINED_ONLY_FOR_INTEGERS, this);
    if (right->type == Type::Long) {
      code.add(op_ldc2w); code.add2(cpool.getLongM1() );
      code.add(op_lxor);
      type = Type::Long;
    } else {
      code.add(op_iconst_m1);
      code.add(op_ixor);
      type = Type::Int;
    }
  }
};

class NotExpression : public UnaryExpression {
public:
  NotExpression(Expression* right):
    UnaryExpression(NOT, Type::Bool, right) {}

  virtual void codeValue() { boolToValue(); }

  virtual void codeBranchIfFalse( int addr ) {
    right->codeBranchIfTrue(addr);
  }

  virtual void codeBranchIfTrue( int addr ) {
    right->codeBranchIfFalse(addr);
  }
};

class NegPosExpression : public UnaryExpression {
public:
  NegPosExpression(int code, Expression* right):
    UnaryExpression(code + POS - ADD, right->type, right){}

  virtual void codeValue() {
    right->codeValue();
    type = right->type;
    code.add(op_ineg + type->offset());
  }
};

class NullExpression : public ConstantExpression {
public:
  NullExpression():ConstantExpression(NULL_E, Type::Null){}
  virtual void codeValue() { code.add(op_aconst_null); }
  virtual void print(ostream& o) const { o << "null"; }
};

class NewInstanceExpression : public NaryExpression {
public:
  NewInstanceExpression(Type* typ, ExprList* args):
    NaryExpression(NEWINSTANCE, typ, 0, args) {}
 
  void codeValue() {
    code.add(op_new).class_type(type);
    code.add(op_dup);
    ClassCtx* ctx = type->getClassCtx();
    MethodCtx* method = asMethodCtx(ctx->safeLocalFind(Identifier::ctor));
    MethodExpression::call(method, args, args_count);
  }

  virtual void print(ostream& o) const {
    o << '(' << opNames[op] << ' ' << type;
    for(Expression* e = args; e; e = e->nextInOwner) o << ' ' << e;
    o << ')';
  }
};

class ShortExpression : public IntegerExpression {
public:
  ShortExpression(short val):
    IntegerExpression(SHORTVAL, Type::Short, val){}
  virtual void print(ostream& o) const {
    o << value << 's';}
};

class BinaryShiftExpression : public BinaryExpression {
public:
  BinaryShiftExpression(int op, Expression* left, Expression* right):
    BinaryExpression(op, left->type, left, right){}

  virtual void codeValue() {
    left->codeValue();
    right->codeValue();

    type = left->type;
    int L = (type == Type::Long)?1:0;
    code.add(op_ishl + L + (op - LSHIFT)*2);
  }

};

class StringExpression : public ConstantExpression {
  Identifier value;
public:
  StringExpression(char * val):
    ConstantExpression(STRINGVAL, Type::String),value(val){}

  virtual bool isString() { return true; }

  virtual void codeValue() {
    u2 id = cpool.addString(value.no());
    if(id < 256) code.add(op_ldc1).byte(id);
            else code.add(op_ldc2).word(id);
  }
  
  //	asm.add(opc_ldc, this);
  virtual void print(ostream& o) const {
    o << '"' << value << '"'; }
};

class SuperExpression : public Expression {
//   LocalField field;
public:
  SuperExpression(Type* type):Expression(SUPER, type){}
  virtual bool isSuper() { return true; }
  virtual MethodCtx * prepareMethodCall() {
    code.add(op_aload_0);
    return asMethodCtx( type->getClassCtx()->localFind(Identifier::ctor) );
  }
};

class ThisExpression : public Expression {
//  LocalField field;
public:
  ThisExpression(Type* type):  Expression(THIS, type){}
  virtual void print(ostream& o) const {
    o << "this";
  }
  virtual MethodCtx * prepareMethodCall() {
    code.add(op_aload_0);
    return asMethodCtx( type->getClassCtx()->localFind(Identifier::ctor) );
  }
};

// class TypeExpression : public Expression {
// public:
//   TypeExpression(Type* type):Expression(TYPE, type){}
//   virtual Identifier toIdent() const {return type->typeSig; }

//   virtual void print(ostream& o) const { o << type; }
// };

class AssignOpExpression : public BinaryExpression {
public:
  AssignOpExpression(int op, Expression* left, Expression* right):
    BinaryExpression(op, Type::Error, left, right) {}

  void assOpCode(bool valNeeded) {
    int depth = left->codeLValue();
    codeDup(depth, 0);
    left->codeLoad();
    type = left->type;

    right->codeValue();
    if(!(op >= ASGLSHIFT && op <= ASGURSHIFT))
      ASSERT2(left->type == right->type, ER_SORRY_NOPROMOTIONS, left->type, right->type);

    if(op >= ASGADD && op <= ASGREM) 
      code.add( op_iadd + 4*(op-ASGADD) + type->offset() );
    else if (op >= ASGLSHIFT && op <= ASGURSHIFT) 
      code.add( op_ishl + 2*(op-ASGLSHIFT) + type->offset() );
    else if (op >= ASGBITAND && op <= ASGBITXOR) 
      code.add( op_iand + 2*(op-ASGBITAND) + type->offset() );
    else
      assert(false);

    if (valNeeded) codeDup(type->stackSize(), depth);
    left->codeStore();
  }
    
  virtual void codeValue() { assOpCode(true); }
  virtual void codeVoid()  { assOpCode(false); }
};
