// sq yearn coy

class JavaStream {
public:
  virtual void skip(u4 size) = 0;
  virtual void read(void* buf, u4 size) = 0;
  float readFloat() { float f; assert(sizeof(f)==4); read(&f,4); return f; }
  int read1() { u1 a; read(&a, 1); return a; }
  u2  read2() { u1 a[2]; read(a,2); return a[0] * 256 + a[1]; }
  u4  read4() { u1 a[4]; read(a,4);
    return 0x10000ul * (a[0] * 256ul + a[1]) + a[2] * 256ul + a[3];  }
  char* readStr(StringAllocator& stringAllocator) {
    u2 len = read2();
    char* ptr = stringAllocator.newStr(len+3);
    ptr[0] = (len >> 8);
    ptr[1] = (len & 255); ptr += 2;
    read(ptr, len); ptr[len] = 0;
    return ptr;
  }
  //  friend JavaStream& operator >> (JavaStream& i, u1& d) {
  //    d = i.read1(); return i; }
  friend JavaStream& operator >> (JavaStream& i, u2& d) {
    d = i.read2(); return i; }
  friend JavaStream& operator >> (JavaStream& i, u4& d) {
    d = i.read4(); return i; } 
};

class FileJavaStream : public JavaStream {
  int size;
  int pos;
  char * fjs_buf;
public:
  explicit FileJavaStream(const char* fname): pos(0), size(0)
  {
    fjs_buf = buf4.mem();
    int fd = open(fname, O_RDONLY);
    assert(fd != -1);
    size = lseek(fd, 0, SEEK_END);
    lseek(fd, 0, SEEK_SET);
    ::read(fd, fjs_buf, size);
    close(fd);
  }
  ~FileJavaStream(){ buf4.release(); }
  virtual void read(void* buf, u4 size) 
    { assert(pos + size < buf4.len);
      memcpy(buf, fjs_buf+pos, size); pos += size; }
  virtual void skip(u4 size) { pos += size; }
};

class FileWriter {
  int pos;
  char * fjs_buf;
  const char* fname;
public:
  explicit FileWriter(const char* name): fname(name), pos(0)
  {
    fjs_buf = buf4.mem();
  }
  ~FileWriter(){ 
    int fd = open(fname, O_WRONLY|O_CREAT);
    assert(fd != -1);
    
    ::write(fd, fjs_buf, pos);
    
    close(fd);
    buf4.release(); 
  }
  void write (void* buf, u4 size) 
    { assert(pos + size < buf4.len);
      memcpy(fjs_buf+pos, buf, size); pos += size; }
};

/*
class FileJavaStream : public JavaStream {
  ifstream i;
public:
  explicit FileJavaStream(const char* fname):i(fname){//,ios::binary){
    assert(!i.bad()); }
  virtual void read(void* buf, u4 size) { i.read(buf, size); }
  virtual void skip(u4 size) { i.ignore(size); }
};
*/
/*
class FileJavaStream : public JavaStream {
  FILE * f;
public:
  explicit FileJavaStream(const char* fname):f(fopen(fname,"rb")){//,ios::binary){
    assert(f); }
  virtual void read(void* buf, u4 size) { fread(buf, size, 1, f);} // cerr << '{' << size << '}'; }
  virtual void skip(u4 size) { fseek(f, SEEK_CUR, size); }
};
*/
class BinaryConstantPool : public ConstantPool{
  StringAllocator stringAllocator;
public:
  BinaryConstantPool():ConstantPool(buf2),stringAllocator(4096){}

  void load(JavaStream& i) {
    int count = i.read2();
    //    cerr << "Loading " << count << " cpool items" << endl;
    reserve(count);
    for(int current = 1; current < count; ++current) {
      int type = i.read1();
      u2 class_ndx, name_utf;
      u4 hi, lo;
      //      cerr << current << ' ' << constant_pool_count << ':' << cpool_type(type) << endl;
      switch( type ) {
        case CONSTANT_Class:      addClass(i.read2()); break;

        case CONSTANT_Fieldref:
        case CONSTANT_Methodref:
        case CONSTANT_IMethodref: class_ndx = i.read2();
                                  addFieldRef(type, class_ndx, i.read2()); break;

        case CONSTANT_String:     addString(i.read2()); break;

        case CONSTANT_Integer:    addInteger(i.read4()); break;
        case CONSTANT_Float:      addFloat(i.readFloat()); break;

        case CONSTANT_Long:
        case CONSTANT_Double:   i >> hi >> lo; addTwoWords(type, hi, lo); ++current; break;

        case CONSTANT_NameAndType:
                                i >> name_utf; addNameAndType(name_utf, i.read2()); break;

        case CONSTANT_Utf8:
				   addUtf(i.readStr(stringAllocator));
				   break;

      }
    }
    prepareAttributeNames();
  }
};

class BinaryClass {
public:
  static u2 constant_value_index; // bad, I know
  JavaStream* in;
  bool readAttr;
  static BinaryClass* rdr;
  BinaryConstantPool cpool;
  void printCode(ostream& cout, const u1* code);
private:

  class BinaryAttribute {
    u2 attribute_name;
    u4 attribute_length;
    u1 * info;
  public:
    BinaryAttribute():info(0),attribute_length(0){assert(false);
      *(BinaryClass::rdr->in) >> attribute_name >> attribute_length;
      if(attribute_length) info = new u1[attribute_length];
      (BinaryClass::rdr->in)->read(info, attribute_length);
      if(attribute_length >= 2 
        && attribute_name == BinaryClass::rdr->cpool.ConstantValue) 
	{
	  constant_value_index = int(info[0])*256 + int(info[1]);
	}
    }
    static void skip(int count) { 
      for(;--count >= 0;) {
        u2 attribute_name; u4 attribute_length;
        *(BinaryClass::rdr->in) >> attribute_name >> attribute_length;
	if(attribute_length >= 2 
        && attribute_name == BinaryClass::rdr->cpool.ConstantValue) 
	{
	  constant_value_index = BinaryClass::rdr->in->read2();
	  attribute_length -= 2;
	}
        (BinaryClass::rdr->in)->skip(attribute_length);
      }
    }
    ~BinaryAttribute(){delete[] info;}
    void print(ostream& cout) {
      BinaryClass::rdr->cpool.printUtf(cout, attribute_name);
      cout << '(' << attribute_length << ") ";
      if(BinaryClass::rdr->cpool.compareUtfWith(attribute_name, "Code"))
        BinaryClass::rdr->printCode(cout, info);
      else if(BinaryClass::rdr->cpool.compareUtfWith(attribute_name, "ConstantValue") ) 
        BinaryClass::rdr->cpool.printConst(cout, int(info[0])*256+int(info[1]));
    }
  };

  class Attributes {
    u2 attributes_count;
    BinaryAttribute * attributes;
  public:
    Attributes():attributes(0),attributes_count(0){}
    void load(){
      *(BinaryClass::rdr->in) >> attributes_count;
      //      cerr << " attr count = " << attributes_count << endl;
      if(BinaryClass::rdr->readAttr && attributes_count)
        attributes = new BinaryAttribute[attributes_count];
      else {
        BinaryAttribute::skip(attributes_count); attributes_count = 0; }
      assert(attributes == 0);
    }
    ~Attributes() {} // dlete[] attributes; }
    void print(ostream& cout) {
      for(int i = 0; i < attributes_count; ++i) attributes[i].print(cout); }
  };

  class FieldInfo {
    u2 access_flags;
    u2 name_index;
    u2 descriptor_index;

    u2 constant_value_index;
    Attributes attributes;
  public:
    FieldInfo() {
      BinaryClass::constant_value_index = 0;
      *(BinaryClass::rdr->in) >> access_flags >> name_index >> descriptor_index;
      attributes.load();
      constant_value_index = BinaryClass::constant_value_index;
    }
    void saveToClassCtx(BinaryConstantPool& cpool, ClassCtx& ctx) {
      ctx.addField(access_flags, 
		   Identifier(cpool.getUtf(name_index)), 
		   Type::typeFromSig(Identifier(cpool.getUtf(descriptor_index))) );
    }
    void print(ostream& cout) {
      printMod(cout, access_flags);
      BinaryClass::rdr->cpool.printUtf(cout, name_index); cout << ':';
      BinaryClass::rdr->cpool.printUtf(cout, descriptor_index);
      cout << ' '; attributes.print(cout);
      if(constant_value_index) {
	cout << " = "; 
	BinaryClass::rdr->cpool.printConst(cout, constant_value_index);
      }
    }
  };

  class Fields {
    u2 fields_count;
    FieldInfo * fields;
  public:
    Fields():fields(0){}
    ~Fields(){delete[] fields;}
    //    ~Fields(){buf5.release();}
    void load() {
      *(BinaryClass::rdr->in) >> fields_count;
      if(fields_count) fields = new FieldInfo[ fields_count ];
    }
    void saveToClassCtx(BinaryConstantPool& cpool, ClassCtx& ctx) {
      for(int i = 0; i < fields_count; ++i) fields[i].saveToClassCtx(cpool, ctx);
    }
    void print(ostream& cout) {
      for(int i = 0; i < fields_count; ++i)
        { fields[i].print(cout); cout << endl; }
    }
  };
  u2 minor, major;
  u2 access_flags;
  u2 this_class;
  u2 super_class;
  u2 interfaces_count;
  u2 * interfaces; // f_info, m_info, a_info
  Fields fields;
  Fields methods;
  Attributes attributes;

public:
  ~BinaryClass() { buf3.release(); } 
  //  ~BinaryClass() { elete[] interfaces; }
  void load(JavaStream& in) {
    this->in = &in;
    rdr = this;
    u4 magic; in >> magic;
    assert(magic == 0xCAFEBABEl);
    in >> minor >> major;
    cpool.load(in);
    in >> access_flags;
    in >> this_class;
    in >> super_class;

    in >> interfaces_count;
    interfaces = (u2*)buf3.mem();
    //    if(interfaces_count) interfaces = new u2[interfaces_count];

    for(int i = 0; i < interfaces_count; ++i) { in >> interfaces[i]; }

    fields.load();
    methods.load();
    attributes.load();
  }

  void saveToClassCtx(ClassCtx& ctx) {
    if(super_class) 
      ctx.setSuperAndFlags(Identifier(cpool.getClassName(super_class)), access_flags);
    else
      ctx.setSuperAndFlags(access_flags);

    for(int j = 0; j < interfaces_count; j++)
      ctx.addInterface( Identifier(cpool.getClassName(interfaces[j])) );
    fields.saveToClassCtx(cpool, ctx);
    methods.saveToClassCtx(cpool, ctx);
  }

  void print(ostream& cout) {
      cout << "Version: " << major << '.' << minor << endl;
    cpool.print(cout);
    printMod(cout, access_flags);
    cout << "class "; cpool.printClass(cout, this_class);
    cout << " extends "; if(super_class) cpool.printClass(cout, super_class);
    cout << " implements";

    for(int j = 0; j < interfaces_count; j++)
      { cout << ' '; cpool.printClass(cout, interfaces[j]); }
    cout << endl;

    fields.print(cout);
    methods.print(cout);
    attributes.print(cout);
    cout << endl;
  }

  explicit BinaryClass(const char * fname, bool ReadAttr):
    interfaces(0),readAttr(ReadAttr)
  { char buf[1024];
    strcpy(buf, fname);
    strcat(buf, ".class");
    FileJavaStream in(buf);
    load(in);
  }
  explicit BinaryClass(JavaStream* stream, bool ReadAttr = true):
    interfaces(0),readAttr(ReadAttr) { load(*in); }
private:
  int Word(const u1* code, int current, int ndx) {
    current += 2 * ndx;
    return 256 * code[current+0] + code[current+1];
  }
  long Long(const u1* code, int current, int ndx) {
    current += 4 * ndx;
    return 65536l * (256 * code[current+0] + code[current+1]) +
                     256 * code[current+2] + code[current+3];
  }
  void printOffs(ostream& o, int current, int offset) {
    o << current + offset << '(';
    if(offset > 0) o << '+';
    o << offset << ')';
  }
};
BinaryClass* BinaryClass::rdr;
u2 BinaryClass::constant_value_index; 

void BinaryClass::printCode(ostream& cout, const u1* code) {
  cout << "Max Stack: " << Word(code,0,0)
       << " Max Locals: " << Word(code,0,1);

  int size = Long(code,0,1);
  code += 8;
  cout << endl;
  for(int current = 0; current < size;) {
    u1 opCode = code[current]; int i;
    assert(opCode <= last_instruction);
    ins_desc ins = ins_descriptions[opCode];
    int b; int cur = current;
    short u;

    cout << setw(4) << current << " : ";

    if(ins.len >= 1) b = (int)code[current+1];
    if(ins.len >= 2) u = 256 * b + code[current+2];
    if(ins.len == 255) current = (current + 4) & ~3;

    if(ins.mode == def_mode) ins.mode = (OpMode)(ins.mode + ins.len);
    cout << ins.name << ' ';

    switch(ins.mode) {
     case u1_lvar: cout << "local_" << b; break;
     case u2_offs: printOffs(cout, current, u); break;
     case u2_class: cpool.printClass(cout, u); break;
     case u1_newarray: printArrCode(cout, code[current+1]); break;
     case interface_mode: cpool.printRef(cout, CONSTANT_IMethodref, u);
                          cout << " nargs=" << (int)code[current+3];
                          break;
     case mref_mode: cpool.printRef(cout, CONSTANT_Methodref, u); break;
     case fref_mode: cpool.printRef(cout, CONSTANT_Fieldref, u); break;
     case iinc_mode: cout << "local_" << b << " += " << (int)(signed char)code[current+2];
                     break;
     case u2_const_ndx: cout << "#" << u <<' '; cpool.printConst(cout, u); break;
     case u1_const_ndx: cout << "#" << b <<' '; cpool.printConst(cout, b); break;
     case class_dim: cout << int(code[current+3]) << ' '; cpool.printClass(cout, u); break;
     case u2_imm: cout << "int_const_" << u; break;
     case u1_imm: cout << "int_const_" << (int)(signed char)b; break;
     case tableswitch_mode: {
       long def_offset = Long(code, current, 0);
       long low = Long(code, current, 1);
       long high = Long(code, current, 2);
       current += 3 * 4;
       cout << "def " << def_offset << ' ' << low << " .. " << high << endl;
       for(i = 0; i < high - low; ++i) {
         short u = 256 * code[current+0] + code[current+1];
         cout << " case " << i + low << " : ";
         printOffs(cout, cur, u); cout << endl;
         current += 2;
       }
       continue; }

    case lookupswitch_mode: {
       long def_offset = Long(code, current, 0);
       long num   = Long(code, current, 1);
       cout << "def " << endl;

       current += 4 * 2;

       for(i = 0; i < num; ++i) {
         long ofs = Long(code, current, 1);
         cout << "  case " << Long(code, current, 0) << " : ";
         printOffs(cout, cur, ofs);
         cout << endl;

         current += 4 * 2;
       }
       continue; }
    }
    current += ins.len + 1;
    cout << endl;
  }
}

void ReadBinaryClass(const char* name, ClassCtx* ctx) {
  cout << "  opening binary " << name << " ... " << flush;
  BinaryClass b(name, false);
  //  b.print(cerr);
  b.saveToClassCtx(*ctx);
  cout << "done" << endl;
  //  cerr << "done loading of " << name << endl;
}
