/************************************************
  Team: Jon Gregory, Gor Nishanov
  Project #2 Part 2
    for RPI course 66-648 Compiler Design Theory

  type.h:

     TypeList class

     Type class has protected constructor.

     To create new types use:

       Type *Type::NewArray;
       Type *Type::NewClass;
       Type *Type::NewMethod;

     or predeclared types

      Type::Char, ... Type::Void

  *****/

class Type;
class ClassCtx;
ClassCtx* globalCtx_findClass(Identifier clazzName);
class Parameter;

struct TypeListItem : public QuickNew {
  Type* type;
  TypeListItem* nextInOwner;
  TypeListItem(Type* typ):type(typ),nextInOwner(0){}
};

struct TypeList : public QuickNew{
  typedef TypeListItem T;
  T* first;
  T* last;
  int count;
  public:
  TypeList():first(0),last(0),count(0){}
  TypeList(T* i):first(i),last(i),count(1){}
  int getCount() const {return count;}
  void add(Type* t) { 
    TypeListItem* i = new TypeListItem(t); ++count;
    if(first == 0) last = first = i;
    else last = (last->nextInOwner = i);
  }
  void clear() { first = last = 0; count = 0; }
  void print(ostream& o) const;
  friend ostream& operator << (ostream& o, const TypeList& t){
    t.print(o); return o; }
  friend ostream& operator << (ostream& o, const TypeList* t){
    t->print(o); return o;}
};

class Type : public QuickNew {
protected:
  static Hash<Identifier, Type*, TYPE_HASH_SIZE> hash;
  static TypeList empty_type_list;

  mutable u2 cpool_no;
  Type(const int typCode, const char* typSig):typeCode(typCode),typeSig(typSig), cpool_no(0) {
    hash.get(typeSig).data = this;
  }

public:
  const int typeCode;
  const Identifier typeSig;
  virtual bool isArray() const { return false; }

  friend ostream& operator << (ostream& o, Type* t) {
    return o << t->getTypeSig() ; }

  static Type *Error, *Null, *Void, *Bool, *Byte, *Char, *Short, *Int,
              *Float, *Long, *Double, *Object, *String, *Cloneable, *StringBuffer;

  const char* getTypeSig() const {return typeSig.asString();}
  int getTypeMask() const {return 1 << typeCode;}
  bool isType(int typCode) const {return typeCode == typCode;}
  bool inMask(int typeMask) const {	return ((1 << typeCode) & typeMask) != 0;	}
  char* copySignatureTo(char* a) const {return typeSig.copyTo(a);}

  static Type* NewArray(Type* elem);
  static Type* NewClass(Identifier className, const ClassCtx* ctx = 0);
  static Type* NewMethod(Type* ret);
  static Type* NewMethod(Type* returnType, Parameter* args);
  static Type* NewMethod(Type* returnType, TypeList* args);

  int no() const { return typeSig.no(); }
  virtual u2 classNo() const { ASSERT(false, ER_NOT_A_CLASS_TYPE, typeSig); return 0; }

  Type* dims(int n) {
    Type* result = this;
    while(--n >= 0) result = NewArray(result);
    return result;
  }

virtual Type* getElementType() const {	
  assert(false /* not an Array */); return Error; }

virtual int getArrayDimension() const {	return 0;}

virtual Identifier getClassName() const {
   assert(false /* not a Class */);  return Identifier::star; }

virtual ClassCtx* getClassCtx() const {
  ASSERT(false, ER_NOT_A_CLASS, this);  return 0; }

virtual Type* getReturnType() const {	
  assert(false /* not a method */);  return Error; }

virtual TypeList* getArgumentTypes() const {
  assert(false /* not a method */); return &empty_type_list; }

virtual bool equalArguments(Type*) const {return false;}

friend ostream& operator << (ostream& o, const Type* t) {
  return o << t->getTypeSig(); }

  virtual bool isMethod() const { return false; }

virtual int stackSize() const {
  switch (typeCode) {
    case TC_ERROR: case TC_VOID: return 0;
    case TC_BOOL:case TC_BYTE:case TC_SHORT:case TC_CHAR:
    case TC_INT:case TC_FLOAT:case TC_ARRAY:case TC_CLASS:return 1;
    case TC_LONG:case TC_DOUBLE:return 2;
  }
  cout << typeCode << endl;
  assert(false /*Bad Type Code */);
  return 0;
}

virtual int arrayCode() const {
  switch (typeCode) {
  case TC_BOOL: return 4;
  case TC_CHAR: return 5;
  case TC_FLOAT: return 6;
  case TC_DOUBLE:return 7;
  case TC_BYTE:  return 8;
  case TC_SHORT: return 9;
  case TC_INT: return 10;
  case TC_LONG: return 11;
  }
  ASSERT2(false, BAD_TYPE_CODE, typeCode, this);
  return 0;
}

int offset() const {
  switch (typeCode) {
    case TC_BOOL:case TC_BYTE:case TC_SHORT:case TC_CHAR:case TC_INT:return 0;
    case TC_LONG:return 1;
    case TC_FLOAT:return 2;
    case TC_DOUBLE:return 3;
    case TC_NULL:case TC_ARRAY:case TC_CLASS:return 4;
  }
  ASSERT2(0, ER_BAD_TYPECODE, typeCode, typeSig);
  return 0;
}

int aoffset() const {
  switch (typeCode) {
    case TC_INT:return 0;
    case TC_LONG:return 1;
    case TC_FLOAT:return 2;
    case TC_DOUBLE:return 3;
    case TC_NULL:case TC_ARRAY:case TC_CLASS:return 4;
    case TC_BOOL:case TC_BYTE: return 5;
    case TC_CHAR: return 6;
    case TC_SHORT: return 7;
  }
  ASSERT2(0, ER_BAD_TYPECODE, typeCode, typeSig);
  return 0;
}

int voffset() const {
  switch (typeCode) {
    case TC_BOOL:case TC_BYTE:case TC_SHORT:case TC_CHAR:case TC_INT:return 0;
    case TC_LONG:return 1;
    case TC_FLOAT:return 2;
    case TC_DOUBLE:return 3;
    case TC_NULL:case TC_ARRAY:case TC_CLASS:return 4;
    case TC_VOID: return 5;
  }
  ASSERT2(0, ER_BAD_TYPECODE, typeCode, typeSig);
  return 0;
}

static Type* typeFromSig(Identifier typeSig) {
  Type*& data = hash.get(typeSig).data;
  if( data ) return data;
  const char* sig = typeSig.asString();
  //    cerr << "<new " << sig << '>';
  
  switch (sig[0]) {
  case SIG_ARRAY: return NewArray(typeFromSig(Identifier(sig+1)) );
    
  case SIG_CLASSBEG:
    return NewClass(Identifier(sig+1, typeSig.len()-2)); // strip 'L' and ';' //
							      
  case SIG_PARBEG: 
    if(sig[1] == SIG_PAREND) return NewMethod(typeFromSig(Identifier(sig+2)));
    TypeList* args = new TypeList();
    sig += 1;
        while(*sig != SIG_PAREND) {
          int j = 0; for(; sig[j] == SIG_ARRAY; ++j);
          if(sig[j] == SIG_CLASSBEG) while( sig[++j] != SIG_CLASSEND );
//          cerr << "[";
//          for(int k = 0; k < j; ++k) cerr << sig[k]; cerr << ']';
          args->add( typeFromSig(Identifier(sig, ++j) ) );
          sig += j;
        }
        return NewMethod(typeFromSig(Identifier(sig+1)), args);  
    }
    cerr << typeSig.asString() << endl;
    ASSERT(false, ER_BAD_TYPESIG, typeSig);
    return Type::Error;
  } 
};

void TypeList::print(ostream& o) const {
  T* f = first;
  if( f == 0 ) o << "<nothing>";
  else while(f != 0) {
    if(f != first) o << ' ';
    o << f->type;
    f = f->nextInOwner;
  }
};


Hash<Identifier, Type*, TYPE_HASH_SIZE> Type::hash;
TypeList Type::empty_type_list;

inline const char* strChar(char ch) {
  static char a[2] = {0,0};
  a[0] = ch;
  return a;
}

class ArrayType : public Type {
  Type * elemType;
  friend Type;
  mutable u2 class_no;
  ArrayType(char* typSig, Type * elType):
    Type(TC_ARRAY,typSig),elemType(elType),class_no(0) {}
public:
  virtual u2 classNo() const { 
    if(!class_no) class_no = cpool.addClass(typeSig.no());
    return class_no;
  }
  virtual bool isArray() const { return true; }
  virtual Type * getElementType() const { return elemType; }
  virtual int getArrayDimension() const { return elemType->getArrayDimension() + 1; }
};

class ClassType : public Type {
  Identifier className; // !! Should be fully qualified
  friend Type;
  ClassType(char* typSig, Identifier clazzName, const ClassCtx * clazz):class_no(0),
    Type(TC_CLASS, typSig),className(clazzName),ctx(const_cast<ClassCtx*>(clazz)){}

  mutable ClassCtx * ctx; 
  mutable u2 class_no;
public:
  virtual u2 classNo() const { 
    if(!class_no) class_no = cpool.addClass(className.no());
    return class_no;
  }
  virtual Identifier getClassName() const {return className;}
  virtual ClassCtx * getClassCtx() const {
    if(!ctx) ctx = globalCtx_findClass(className);
    return ctx;
  }
};

class MethodType : public Type {
  Type * returnType;
  TypeList * argTypes;
  friend Type;
  MethodType(char* typSig, Type* retType, TypeList* argTyps):
    Type(TC_METHOD, typSig), returnType(retType), argTypes(argTyps) {}
public:
  virtual bool isMethod() const { return true; }
  virtual Type * getReturnType() const {return returnType;}
  virtual TypeList * getArgumentTypes() const {	return argTypes; }
  virtual bool equalArguments(Type* t) const {
    if (!t->isType(TC_METHOD)) return false;
    MethodType* m = (MethodType*)t;
    if (argTypes->getCount() != m->argTypes->getCount()) return false;
    TypeListItem * i = argTypes->first, *j = m->argTypes->first;
    while(i != 0) {
      if(i->type != j->type) return false;
      i = i->nextInOwner; j = j->nextInOwner;
    }
    return true;
  }
  virtual int stackSize() const {
    int n = 0;
    TypeListItem* i = argTypes->first;
    while(i != 0) {
      n += i->type->stackSize();
      i = i->nextInOwner;
    }
    return n;
  }
};

////////////////
Type* Type::NewArray(Type* elem) {
  char sig[255]; sig[0] = SIG_ARRAY; elem->typeSig.copyTo(sig+1);
  Type** t = &(hash.get(Identifier(sig)).data);
  if(*t == 0) *t = new ArrayType(sig, elem);
  return *t;
}

Type * Type::NewClass(Identifier className, const ClassCtx* ctx) {
  char sig[255]; sig[0] = SIG_CLASSBEG;
  char * end = className.copyTo(sig+1);	end[0] = SIG_CLASSEND; end[1] = 0;
  //  cerr << "Type::NewClass(" << sig << ')' << endl;;
  Type** t = &(hash.get(Identifier(sig)).data);
  if(*t == 0) *t = new ClassType(sig, className, ctx);
  return *t;
}

class Parameter : public QuickNew {
public:
  Type * const    type;
  const Identifier name;

  Parameter(Identifier aName, Type* aType): name(aName), type(aType) {}

  friend ostream& operator << (ostream& o, const Parameter* p) {
     return o << p->type << " " << p->name; 
  }

  TypeList * NewTypeList() {
    TypeList * list = new TypeList();
    Parameter* p = this;
    while(p) {
      list->add( p->type );
      p = p -> nextInOwner;
    }
    return list;
  }

  Parameter * nextInOwner;
};

Type* Type::NewMethod(Type* returnType, Parameter * args) {
  char sig[1024]; // !!! Later we should use get_tmp or sth
  char * ptr = sig+1; sig[0] = SIG_PARBEG;
  Parameter * p = args;
  while(p) {
    ptr = p->type->copySignatureTo(ptr);
    p = p -> nextInOwner;
  }
  *ptr = SIG_PAREND; ptr = returnType->copySignatureTo(++ptr); *ptr = 0;
  Type** t = &(hash.get(Identifier(sig)).data);
  if(*t == 0) {
    *t = new MethodType(sig, returnType, args?args->NewTypeList():&empty_type_list );
  }
  return *t;
}
Type* Type::NewMethod(Type* ret) {
  return NewMethod(ret, &empty_type_list);
}

Type* Type::NewMethod(Type* returnType, TypeList* argTypes) {
  char sig[1024]; // !!! Later we should use get_tmp or sth
  char * ptr = sig+1; sig[0] = SIG_PARBEG;
  TypeListItem * i = argTypes->first;
  while(i != NULL) {
    ptr = i->type->copySignatureTo(ptr);
    i = i -> nextInOwner;
  }
  *ptr = SIG_PAREND; ptr = returnType->copySignatureTo(++ptr); *ptr = 0;
  Type** t = &(hash.get(Identifier(sig)).data);
  if(*t == 0) *t = new MethodType(sig, returnType, argTypes);
  return *t;
}

MethodType* asMethodType(Type* t) {
  ASSERT(t->isMethod(), ER_SHOULD_BE_METHOD_TYPE, t);
  return (MethodType*) t;
}

Type * Type::Error 	  = new Type(TC_ERROR,   "?");
Type * Type::Null 	  = new Type(TC_NULL, 	  "*");
Type * Type::Void 	  = new Type(TC_VOID, 	 strChar(SIG_VOID) );
Type * Type::Bool    	  = new Type(TC_BOOL, 	 strChar(SIG_BOOL) );
Type * Type::Byte 	  = new Type(TC_BYTE, 	 strChar(SIG_BYTE) );
Type * Type::Char 	  = new Type(TC_CHAR, 	 strChar(SIG_CHAR) );
Type * Type::Short 	  = new Type(TC_SHORT,	 strChar(SIG_SHORT) );
Type * Type::Int          = new Type(TC_INT,    strChar(SIG_INT) ); 
Type * Type::Float 	  = new Type(TC_FLOAT,	 strChar(SIG_FLOAT) );
Type * Type::Long 	  = new Type(TC_LONG, 	 strChar(SIG_LONG) );
Type * Type::Double 	  = new Type(TC_DOUBLE, strChar(SIG_DOUBLE) );
Type * Type::Object 	  = Type::NewClass(Identifier::JavaLangObject);
Type * Type::String 	  = Type::NewClass(Identifier::JavaLangString);
Type * Type::StringBuffer = Type::NewClass(Identifier::JavaLangStringBuffer);
Type * Type::Cloneable    = Type::NewClass(Identifier::JavaLangCloneable);
////////////////

